import contextlib
import pathlib
import numpy as np
import pytest
import sxs


def test_table():
    """Ensure that we can compare floats as expected"""
    with contextlib.redirect_stdout(None):
        catalog = sxs.load("catalog")
    table = catalog.table
    assert len(table[table["object_types"] == "BHBH"]) >= 2019
    assert len(table[table["object_types"] == "BHNS"]) >= 7
    assert len(table[table["object_types"] == "NSNS"]) >= 2
    for name, dtype in zip(table.columns, table.dtypes):
        if any(name.startswith(s) for s in ["initial_", "com_correction_", "reference_"]):
            if name == "initial_data_type":
                continue
            if dtype == np.float64:
                table[table[name] > 1e-7]
            elif dtype == object:
                for i, vec in enumerate(table[name]):
                    assert isinstance(vec, np.ndarray), f"table.iloc[{i}][{name}] = {vec}"
                    assert vec.dtype == np.float64, f"table.iloc[{i}][{name}] = {vec}"
                    assert vec.size == 3, f"table.iloc[{i}][{name}] = {vec}"
            else:
                raise ValueError(f"Illegal dtype {dtype} for column {name}")
    for url in table["url"]:
        assert sxs.utilities.url.validate(url)
    for metadata_path in table["metadata_path"]:
        path = pathlib.Path(metadata_path)  # Ensure that we can do this
        assert sxs.sxs_id(str(path.parts[0]))  # Ensure that this starts with an SXS ID



journal_macros = {
    r'\pra': 'Phys. Rev. A',
    r'\prb': 'Phys. Rev. B',
    r'\prc': 'Phys. Rev. C',
    r'\prd': 'Phys. Rev. D',
    r'\prl': 'Phys. Rev. Lett.',
    r'\prx': 'Phys. Rev. X',
    r'\mnras': 'Mon. Not. R. Astron Soc.',
    r'MNRAS': 'Mon. Not. R. Astron Soc.',
}

# Taken from the "Reviews of Modern Physics Style Guide", Appendix B, with a few revisions
journal_abbreviation_pairs = {
    "Accounts of Chemical Research": "Acc. Chem. Res.",
    "Acta Chemica Scandinavica": "Acta Chem. Scand.",
    "Acta Crystallographica": "Acta Crystallogr.",
    "Acta Crystallographica, Section A: Crystal Physics, Diffraction, Theoretical and General Crystallography": "Acta Crystallogr. Sec. A",
    "Acta Crystallographica, Section B: Structural Crystallography and Crystal Chemistry": "Acta Crystallogr. Sec. B",
    "Acta Mathematica Academiae Scientiarum Hungaricae": "Acta Math. Acad. Sci. Hung.",
    "Acta Metallurgica": "Acta Metall.",
    "Acta Physica": "Acta Phys.",
    "Acta Physica Austriaca": "Acta Phys. Austriaca",
    "Acta Physica Polonica": "Acta Phys. Pol.",
    "Acustica": "Acustica",
    "Advances in Applied Mechanics": "Adv. Appl. Mech.",
    "Advances in Atomic and Molecular Physics": "Adv. At. Mol. Phys.",
    "Advances in Chemical Physics": "Adv. Chem. Phys.",
    "Advances in Magnetic Resonance": "Adv. Magn. Reson.",
    "Advances in Nuclear Physics": "Adv. Nucl. Phys.",
    "Advances in Physics": "Adv. Phys.",
    "Advances in Quantum Chemistry": "Adv. Quantum Chem.",
    "AIAA Journal": "AIAA J.",
    "AIChE Journal": "AIChE J.",
    "AIP Conference Proceedings": "AIP Conf. Proc.",
    "Akusticheskii Zhurnal [Soviet Physics Acoustics]": "Akust. Zh. [Sov. Phys. Acoust.]",
    "American Journal of Physics": "Am. J. Phys.",
    "Analytical Chemistry": "Anal. Chem.",
    "Annalen der Physik (Leipzig)": "Ann. Phys. (Leipzig)",
    "Annales de Chimie et de Physique": "Ann. Chim. Phys.",
    "Annales de Geophysique": "Ann. Geophys.",
    "Annales de l'Institut Henri Poincaré": "Ann. Inst. Henri Poincaré",
    "Annales de l'Institut Henri Poincaré, Section A: Physique Theorique": "Ann. Inst. Henri Poincaré, A",
    "Annales de l'Institut Henri Poincaré, Section B: Calcul des Probabilités et Statistique": "Ann. Inst. Henri Poincaré, B",
    "Annales de Physique (Paris)": "Ann. Phys. (Paris)",
    "Annals of Fluid Dynamics": "Ann. Fluid Dyn.",
    "Annals of Mathematics": "Ann. Math.",
    "Annals of Physics (New York)": "Ann. Phys. (N.Y.)",
    "Annual Review of Astronomy and Astrophysics": "Annu. Rev. Astron. Astrophys.",
    "Annual Review of Atomic and Molecular Physics": "Annu. Rev. At. Mol. Phys.",
    "Annual Review of Fluid Mechanics": "Annu. Rev. Fluid Mech.",
    "Annual Review of Nuclear Science": "Annu. Rev. Nucl. Sci.",
    "Applied Optics": "Appl. Opt.",
    "Applied Physics Letters": "Appl. Phys. Lett.",
    "Applied Spectroscopy": "Appl. Spectrosc.",
    "Arkiv foer Fysik": "Ark. Fys.",
    "Astronomical Journal": "Astron. J.",
    "Astronomicheskii Zhurnal [Soviet Astronomy]": "Astron. Zh. [Sov. Astron.]",
    "Astronomische Nachrichten": "Astron. Nachr.",
    "Astronomy and Astrophysics": "Astron. Astrophys.",
    "Astroparticle Physics": "Astropart. Phys.",
    "Astrophysical Journal": "Astrophys. J.",
    "Astrophysical Journal, Letters to the Editor": "Astrophys. J. Lett.",
    "Astrophysical Journal, Supplement Series": "Astrophys. J. Suppl. Ser.",
    "Astrophysical Letters": "Astrophys. Lett.",
    "Astrophysics and Space Science": "Astrophy. Space Sci.",
    "Atomic Data and Nuclear Data Tables": "At. Data. Nucl. Data Tables",
    "Atomnaya Energiya [Soviet Journal of Atomic Energy]": "At. Energ. [Sov. J. At. Energy]",
    "Australian Journal of Physics": "Aust. J. Phys.",
    "Bell System Technical Journal": "Bell Syst. Tech. J.",
    "Berichte der Bunsengesellschaft für Physikalische Chemie": "Ber. Bunsenges. Phys. Chem.",
    "British Journal of Applied Physics": "Br. J. Appl. Phys.",
    "Bulletin of the Academy of Sciences of the USSR, Physical Series [translation of Izvestiya Akademii Nauk SSSR, Seriya Fizicheskaya]": "Bull. Acad. Sci. USSR, Phys. Ser.",
    "Bulletin of the American Astronomical Society": "Bull. Am. Astron. Soc.",
    "Bulletin of The American Physical Society": "Bull. Am. Phys. Soc.",
    "Bulletin of the Astronomical Institutes of the Netherlands": "Bull. Astron. Inst. Neth.",
    "Bulletin of the Chemical Society of Japan": "Bull. Chem. Soc. Jpn.",
    "Bulletin of the Seismological Society of America": "Bull. Seismol. Soc. Am.",
    "Canadian Journal of Chemistry": "Can. J. Chem.",
    "Canadian Journal of Physics": "Can. J. Phys.",
    "Canadian Journal of Research": "Can. J. Res.",
    "Chaos": "Chaos",
    "Chemical Physics": "Chem. Phys.",
    "Chemical Physics Letters": "Chem. Phys. Lett.",
    "Chemical Reviews": "Chem. Rev.",
    "Chinese Journal of Physics [translation of Wuli Xuebao (Acta Physica Sinica)]": "Chin. J. Phys.",
    "Classical and Quantum Gravity": "Class. Quantum Gravity",
    "Comments on Astrophysics and Space Physics": "Comments Astrophys. Space Phys.",
    "Comments on Atomic and Molecular Physics": "Comments At. Mol. Phys.",
    "Comments on Nuclear and Particle Physics": "Comments Nucl. Part. Phys.",
    "Comments on Plasma Physics and Controlled Fusion": "Comments Plasma Phys. Controlled Fusion",
    "Comments on Solid State Physics": "Comments Solid State Phys.",
    "Communications in Mathematical Physics": "Commun. Math. Phys.",
    "Communications on Pure and Applied Mathematics": "Commun. Pure Appl. Math.",
    "Complex Systems": "Complex Syst.",
    "Comptes Rendus Hebdomadaires des Séances de l'Académie des Sciences": "C. R. Acad. Sci.",
    "Comptes Rendus Hebdomadaires des Séances de l'Académie des Sciences, Serie A: Sciences Mathématiques": "C. R. Acad. Sci. Ser. A",
    "Comptes Rendus Hebdomadaires des Séances de l'Académie des Sciences, Serie B: Sciences Physiques": "C. R. Acad. Sci. Ser. B",
    "Cryogenics": "Cryogenics",
    "Czechoslovak Journal of Physics": "Czech. J. Phys.",
    "Discussions of the Faraday Society": "Discuss. Faraday Soc.",
    "Doklady Akademii Nauk SSSR [Soviet Physics–Doklady]": "Dokl. Akad. Nauk SSSR [Sov. Phys. Dokl.]",
    "Earth and Planetary Science Letters": "Earth Planet. Sci. Lett.",
    "Earth Planets Space [formerly Journal of Geomagnetism  and Geoelectricity]": "Earth Planets Space",
    "Electronics Letters": "Electron. Lett.",
    "European Journal of Physics": "Eur. J. Phys.",
    "European Physical Journal A: Hadrons and Nuclei": "Eur. Phys. J. A",
    "European Physics Journal B: Condensed Matter": "Eur. Phys. J. B",
    "European Physics Journal C: Particles and Fields": "Eur. Phys. J. C",
    "European Physics Journal D: Atomic, Molecular,  and Optical Physics": "Eur. Phys. J. D",
    "European Physics Journal E: Soft Matter": "Eur. Phys. J. E",
    "Europhysics Letters": "Europhys. Lett.",
    "Ferroelectrics": "Ferroelectrics",
    "Few-Body Systems": "Few-Body Syst.",
    "Fields and Quanta": "Fields Quanta",
    "Fizika Elementarnykh Chastits i Atomnogo Yadra [Soviet Journal of  Particles and Nuclei]": "Fiz. Elem. Chastits At. Yadra [Sov. J. Part. Nucl.]",
    "Fizika i Tekhnika Poluprovodnikov [Soviet Physics–Semiconductors]": "Fiz. Tekh. Poluprovodn. [Sov. Phys. Semicond.]",
    "Fizika Metallov i Metallovedenie [Physics of Metals and  Metallography (USSR)]": "Fiz. Met. Metalloved. [Phys. Met. Metallogr. (USSR)]",
    "Fizika Nizkikh Temperatur [Soviet Journal of  Low Temperature Physics]": "Fiz. Nizk. Temp. [Sov. J. Low Temp. Phys.]",
    "Fizika Plazmy [Soviet Journal of  Plasma Physics]": "Fiz. Plazmy [Sov. J. Plasma Phys.]",
    "Fizika Tverdogo Tela (Leningrad) [Soviet Physics–Solid State]": "Fiz. Tverd. Tela (Leningrad) [Sov. Phys. Solid State]",
    "Fortschritte der Physik": "Fortschr. Phys.",
    "Foundations of Physics": "Found. Phys.",
    "General Relativity and Gravitation": "Gen. Relativ. Gravit.",
    "Geochimica et Cosmochimica Acta": "Geochim. Cosmochim. Acta",
    "Helvetica Chimica Acta": "Helv. Chim. Acta",
    "Helvetica Physica Acta": "Helv. Phys. Acta",
    "High Temperature (USSR) [translation of  Teplofizika Vysokikh Temperatur]": "High Temp. (USSR)",
    "Hyperfine Interactions": "Hyperfine Interact.",
    "IBM Journal of Research and Development": "IBM J. Res. Dev.",
    "Icarus": "Icarus",
    "IEEE Journal of Quantum Electronics": "IEEE J. Quantum Electron.",
    "IEEE Transactions on Electron Devices": "IEEE Trans. Electron Devices",
    "IEEE Transactions on Information Theory": "IEEE Trans. Inf. Theory",
    "IEEE Transactions on Instrumentation and Measurement": "IEEE Trans. Instrum. Meas.",
    "IEEE Transactions on Magnetics": "IEEE Trans. Magn.",
    "IEEE Transactions on Microwave Theory and Techniques": "IEEE Trans. Microwave Theory Tech.",
    "IEEE Transactions on Nuclear Science": "IEEE Trans. Nucl. Sci.",
    "IEEE Transactions on Sonics and Ultrasonics": "IEEE Trans. Sonics Ultrason.",
    "Industrial and Engineering Chemistry": "Ind. Eng. Chem.",
    "Infrared Physics": "Infrared Phys.",
    "Inorganic Chemistry": "Inorg. Chem.",
    "Inorganic Materials (USSR) [translation of  Izvestiya Akademii Nauk SSSR, Neorganicheskie Materialy]": "Inorg. Mater. (USSR)",
    "Instruments and Experimental Techniques (USSR) [translation of  Pribory i Tekhnika Eksperimenta]": "Instrum. Exp. Tech. (USSR)",
    "International Journal of Energy Research": "Int. J. Energy Res.",
    "International Journal of Magnetism": "Int. J. Magn.",
    "International Journal of Quantum Chemistry": "Int. J. Quantum Chem.",
    "International Journal of Quantum Chemistry, Part 1": "Int. J. Quantum Chem. 1",
    "International Journal of Quantum Chemistry, Part 2": "Int. J. Quantum Chem. 2",
    "International Journal of Quantum Information": "Int. J. Quantum Inf.",
    "International Journal of Theoretical Physics": "Int. J. Theor. Phys.",
    "Izvestiya, Academy of Sciences, USSR, Atmospheric  and Oceanic Physics [translation of Izvestiya Akademii Nauk SSSR, Fizika Atmosfery i Okeana]": "Izv. Acad. Sci. USSR, Atmos. Oceanic Phys.",
    "Izvestiya, Academy of Sciences, USSR, Physics of  the Solid Earth [translation of  Izvestiya Akademii Nauk SSSR, Fizika Zemli]": "Izv. Acad. Sci. USSR, Phys. Solid Earth",
    "Izvestiya Akademii Nauk SSSR, Fizika": "Izv. Acad. Nauk SSSR, Fiz. Atmos. Okeana",
    "Atmosfery i Okeana [Izvestiya, Academy of Sciences, USSR,  Atmospheric and Oceanic Physics]": "[Izv. Acad. Sci. USSR, Atmos. Oceanic Phys.]",
    "Izvestiya Akademii Nauk SSSR, Fizika Zemli [Izvestiya, Academy of Sciences, USSR, Physics of the Solid Earth]": "Izv. Acad. Nauk SSSR, Fiz. Zemli. [Izv. Acad. Sci. USSR, Phys. Solid Earth]",
    "Izvestiya Akademii Nauk SSSR, Neorganicheskie": "Izv. Akad. Nauk SSSR, Neorg. Mater.",
    "Materialy [Inorganic Materials (USSR)]": "Materialy [Inorg. Mater. (USSR)]",
    "Izvestiya Akademii Nauk SSSR, Seriya Fizicheskaya [Bulletin of the Academy of Sciences of the USSR, Physical Series]": "Izv. Acad. Nauk SSSR, Ser. Fiz. [Bull. Acad. Sci. USSR, Phys. Ser.]",
    "Izvestiya Vysshikh Uchebnykh Zavedenii,  Fizika [Soviet Physics Journal]": "Izv. Vyssh. Uchebn. Zavad. Fiz. [Sov. Phys. J.]",
    "Izvestiya Vysshikh Uchebnykh Zavedenii, Radiofizika [Soviet Radiophysics]": "Izv. Vyssh. Uchebn. Zaved. Radiofiz. [Sov. Radiophys.]",
    "Japanese Journal of Applied Physics": "Jpn. J. Appl. Phys.",
    "Japanese Journal of Physics": "Jpn. J. Phys.",
    "JETP Letters [translation of Pisma v Zhurnal Eksperimentalnoi i Teoreticheskoi Fiziki]": "JETP Lett.",
    "Journal de Chimie Physique": "J. Chim. Phys.",
    "Journal de Physique (Paris)": "J. Phys. (Paris)",
    "Journal de Physique et le Radium": "J. Phys. Radium",
    "Journal of Applied Crystallography": "J. Appl. Crystallogr.",
    "Journal of Applied Physics": "J. Appl. Phys.",
    "Journal of Applied Spectroscopy (USSR) [translation of  Zhurnal Prikladnoi Spektroskopii]": "J. Appl. Spectrosc. (USSR)",
    "Journal of Atmospheric and Terrestrial Physics": "J. Atmos. Terr. Phys.",
    "Journal of Atmospheric Sciences": "J. Atmos. Sci.",
    "Journal of Chemical Physics": "J. Chem. Phys.",
    "Journal of Colloid and Interface Science": "J. Colloid Interface Sci.",
    "Journal of Computational Physics": "J. Comput. Phys.",
    "Journal of Crystal Growth": "J. Cryst. Growth",
    "Journal of Electron Spectroscopy and Related Phenomenon": "J. Electron. Spectrosc. Relat. Phenom.",
    "Journal of Electronic Materials": "J. Electron. Mater.",
    "Journal of Fluid Mechanics": "J. Fluid Mech.",
    "Journal of Geophysical Research": "J. Geophys. Res.",
    "Journal of High Energy Physics": "J. High Energy Phys.",
    "Journal of Inorganic and Nuclear Chemistry": "J. Inorg. Nucl. Chem.",
    "Journal of Low Temperature Physics": "J. Low Temp. Phys.",
    "Journal of Luminescence": "J. Lumin.",
    "Journal of Macromolecular Science, [Part B] Physics": "J. Macromol. Sci. Phys.",
    "Journal of Magnetism and Magnetic Materials": "J. Magn. Magn. Mater.",
    "Journal of Mathematical Physics (New York)": "J. Math. Phys. (N.Y.)",
    "Journal of Molecular Spectroscopy": "J. Mol. Spectrosc.",
    "Journal of Non-Crystalline Solids": "J. Non-Cryst. Solids",
    "Journal of Nonlinear Science": "J. Nonlinear Sci.",
    "Journal of Nuclear Energy": "J. Nucl. Energy",
    "Journal of Nuclear Energy, Part C: Plasma Physics, Accelerators, Thermonuclear Research": "J. Nucl. Energy, Part C",
    "Journal of Nuclear Materials": "J. Nucl. Mater.",
    "Journal of Physical and Chemical Reference Data": "J. Phys. Chem. Ref. Data",
    "Journal of Physical Chemistry": "J. Phys. Chem.",
    "Journal of Physics A: Mathematical and General": "J. Phys. A",
    "Journal of Physics and Chemistry of Solids": "J. Phys. Chem. Solids",
    "Journal of Physics: Condensed Matter": "J. Phys. Condens. Matter",
    "Journal of Physics B: Atomic, Molecular and Optical": "J. Phys. B",
    "Journal of Physics C: Solid State Physics": "J. Phys. C",
    "Journal of Physics D: Applied Physics": "J. Phys. D",
    "Journal of Physics E: Scientific Instruments": "J. Phys. E",
    "Journal of Physics F: Metal Physics": "J. Phys. F",
    "Journal of Physics G: Nuclear and Particle Physics": "J. Phys. G",
    "Journal of Physics (Moscow)": "J. Phys. (Moscow)",
    "Journal of Plasma Physics": "J. Plasma Phys.",
    "Journal of Polymer Science": "J. Polym. Sci.",
    "Journal of Polymer Science, Polymer Letters Edition": "J. Polym. Sci. Polym. Lett. Ed.",
    "Journal of Polymer Science, Polymer Physics Edition": "J. Polym. Sci. Polym. Phys. Ed.",
    "Journal of Quantitative Spectroscopy & Radiative Transfer": "J. Quant. Spectrosc. Radiat. Transfer",
    "Journal of Research of the National Bureau of Standards": "J. Res. Natl. Bur. Stand.",
    "Journal of Research of the National Bureau of Standards, Section A: Physics and Chemistry": "J. Res. Natl. Bur. Stand. Sec. A",
    "Journal of Research of the National Bureau of Standards, Section B: Mathematical Sciences": "J. Res. Natl. Bur. Stand. Sec. B",
    "Journal of Research of the National Bureau of Standards, Section C: Engineering and Instrumentation": "J. Res. Natl. Bur. Stand. Sec. C",
    "Journal of Research of the National Institute of Standards and Technology": "J. Res. Natl. Inst. Stand. Technol.",
    "Journal of Scientific Instruments": "J. Sci. Instrum.",
    "Journal of Sound and Vibration": "J. Sound Vib.",
    "Journal of Statistical Physics": "J. Stat. Phys.",
    "Journal of Superconductivity": "J. Supercond.",
    "Journal of the Acoustical Society of America": "J. Acoust. Soc. Am.",
    "Journal of the American Ceramic Society": "J. Am. Ceram. Soc.",
    "Journal of the American Chemical Society": "J. Am. Chem. Soc.",
    "Journal of the American Institute of Electrical Engineers": "J. Am. Inst. Electr. Eng.",
    "Journal of the Audio Engineering Society": "J. Audio Eng. Soc.",
    "Journal of the Chemical Society": "J. Chem. Soc.",
    "Journal of the Electrochemical Society": "J. Electrochem. Soc.",
    "Journal of the Mechanics and Physics of Solids": "J. Mech. Phys. Solids",
    "Journal of the Optical Society of America": "J. Opt. Soc. Am.",
    "Journal of the Optical Society of America A: Optics, Image Science & Vision": "J. Opt. Soc. Am. A",
    "Journal of the Optical Society of America B: Optical Physics": "J. Opt. Soc. Am. B",
    "Journal of Theoretical Biology": "J. Theor. Biol.",
    "Journal of the Physical Society of Japan": "J. Phys. Soc. Jpn.",
    "Journal of Vacuum Science and Technology": "J. Vac. Sci. Technol.",
    "Kolloid Zeitschrift & Zeitschrift für Polymere": "Kolloid Z. Z. Polym.",
    "Kongelige Danske Videnskabernes Selskab, Matematisk-Fysiske Meddelelser": "K. Dan. Vidensk. Selsk. Mat. Fys. Medd.",
    "Kristallografiya [Soviet Physics–Crystallography]": "Kristallografiya [Sov. Phys. Crystallogr.]",
    "Kvantovaya Elektronika (Moscow) [Soviet Journal of Quantum Electronics]": "Kvant. Elektron. (Moscow) [Sov. J. Quantum Electron.]",
    "Lettere al Nuovo Cimento": "Lett. Nuovo Cimento",
    "Laser and Particle Beams": "Laser Part. Beams",
    "Lick Observatory Bulletin": "Lick Obs. Bull.",
    "Materials Research Bulletin": "Mater. Res. Bull.",
    "Materials Science and Engineering": "Mater. Sci. Eng.",
    "Mathematical Biosciences": "Math. Biosci.",
    "Mathematical Physics and Applied Mathematics": "Math. Phys. Appl. Math.",
    "Medical Physics": "Med. Phys.",
    "Memoirs of the Royal Astronomical Society": "Mem. R. Astron. Soc.",
    "Molecular Crystals and Liquid Crystals": "Mol. Cryst. Liq. Cryst.",
    "Molecular Physics": "Mol. Phys.",
    "Monthly Notices of the Royal Astronomical Society": "Mon. Not. R. Astron. Soc.",
    "National Bureau of Standards (U.S.), Circular": "Natl. Bur. Stand. Circ. (U.S.)",
    "National Bureau of Standards (U.S.), Miscellaneous Publication": "Natl. Bur. Stand. Misc. Publ. (U.S.)",
    "National Bureau of Standards (U.S.), Special Publication": "Natl. Bur. Stand. Spec. Publ. (U.S.)",
    "Nature (London)": "Nature (London)",
    "Naturwissenschaften": "Naturwissenschaften",
    "Nuclear Data, Section A": "Nucl. Data, Sec. A",
    "Nuclear Data, Section B": "Nucl. Data, Sec. B",
    "Nuclear Fusion": "Nucl. Fusion",
    "Nuclear Instruments": "Nucl. Instrum.",
    "Nuclear Instruments & Methods": "Nucl. Instrum. Methods",
    "Nuclear Physics": "Nucl. Phys.",
    "Nuclear Physics A": "Nucl. Phys. A",
    "Nuclear Physics B": "Nucl. Phys. B",
    "Nuclear Science and Engineering": "Nucl. Sci. Eng.",
    "Nukleonika": "Nukleonika",
    "Nuovo Cimento": "Nuovo Cimento",
    "Nuovo Cimento A": "Nuovo Cimento A",
    "Nuovo Cimento B": "Nuovo Cimento B",
    "Optica Acta": "Opt. Acta",
    "Optics and Spectroscopy (USSR) [translation of Optika i Spektroskopiya]": "Opt. Spectrosc.",
    "Optics Communications": "Opt. Commun.",
    "Optics Letters": "Opt. Lett.",
    "Optics News": "Opt. News",
    "Optik (Stuttgart)": "Optik (Stuttgart)",
    "Optika i Spektroskopiya [Optics and Spectroscopy (USSR)]": "Opt. Spektrosk. [Opt. Spectrosc. (USSR)]",
    "Optiko–Mekhanicheskaya Promyshlennost [Soviet Journal of Optical Technology]": "Opt.–Mekh. Prom. [Sov. J. Opt. Technol.]",
    "Phase Transition and Critical Phenomena": "Phase Transit. Crit. Phenom.",
    "Phase Transitions": "Phase Transit.",
    "Philips Research Reports": "Philips Res. Rep.",
    "Philosophical Magazine": "Philos. Mag.",
    "Philosophical Transactions of the Royal Society of London": "Philos. Trans. R. Soc. London",
    "Philosophical Transactions of the Royal Society of London, Series A: Mathematical and Physical Sciences": "Philos. Trans. R. Soc. London, Ser. A",
    "Physica A": "Physica A",
    "Physica B": "Physica B",
    "Physical Chemistry Chemical Physics": "Phys. Chem. Chem. Phys.",
    "Physica Scripta": "Phys. Scr.",
    "Physica Status Solidi": "Phys. Status Solidi",
    "Physica Status Solidi A: Applied Research": "Phys. Status Solidi A",
    "Physica Status Solidi B: Basic Research": "Phys. Status Solidi B",
    "Physica (Utrecht)": "Physica (Utrecht)",
    "Physical Review": "Phys. Rev.",
    "Physical Review A: Atomic, Molecular, and Optical Physics": "Phys. Rev. A",
    "Physical Review B: Condensed Matter": "Phys. Rev. B",
    "Physical Review C: Nuclear Physics": "Phys. Rev. C",
    "Physical Review D: Particles and Fields": "Phys. Rev. D",
    "Physical Review E: Statistical Physics, Plasmas, Fluids, and Related Interdisciplinary Topics": "Phys. Rev. E",
    "Physical Review Letters": "Phys. Rev. Lett.",
    "Physical Review Special Topics – Accelerators and Beams": "Phys. Rev. Spec. Top. Accel. Beams",
    "Physics and Chemistry of Solids": "Phys. Chem. Solids",
    "Physics Letters": "Phys. Lett.",
    "Physics Letters A": "Phys. Lett. A",
    "Physics Letters B": "Phys. Lett. B",
    "Physics (New York)": "Physics (N.Y.)",
    "Physics of Fluids": "Phys. Fluids",
    "Physics of Metals and Metallography (USSR) [translation of Fizika Metallovi Metallovedenie]": "Phys. Met. Metallogr. (USSR)",
    "Physics of Plasmas": "Phys. Plasmas",
    "Physics Reports": "Phys. Rep.",
    "Physics Teacher": "Phys. Teach.",
    "Physics Today": "Phys. Today",
    "Physikalische Zeitschrift": "Phys. Z.",
    "Physikalische Zeitschrift der Sowjetunion": "Phys. Z. Sowjetunion",
    "Physik der Kondensierten Materie": "Phys. Kondens. Mater.",
    "Pis'ma v Astronomicheskii Zhurnal [Soviet Astronomy Letters]": "Pis'ma Astron. Zh. [Sov. Astron. Lett.]",
    "Pis'ma v Zhurnal Eksperimentalnoi i Teoreticheskoi Fiziki [JETP Letters]": "Pis'ma Zh. Eksp. Teor. Fiz. [JETP Lett.]",
    "Pis'ma v Zhurnal Tekhnicheskoi Fiziki [Soviet Technical Physics Letters]": "Pis'ma Zh. Tekh. Fiz. [Sov. Tech. Phys. Lett.]",
    "Planetary and Space Science": "Planet. Space Sci.",
    "Plasma Physics": "Plasma Phys.",
    "Plasma Physics and Controlled Fusion": "Plasma Phys. Control. Fusion",
    "Pribory i Tekhnika Eksperimenta [Instruments and Experimental Techniques (USSR)]": "Prib. Tekh. E ksp. [Instrum. Exp. Tech. (USSR)]",
    "Proceedings of the Cambridge Philosophical Society": "Proc. Cambridge Philos. Soc.",
    "Proceedings of the IEEE": "Proc. IEEE",
    "Proceedings of the IRE": "Proc. IRE",
    "Proceedings of the National Academy of Sciences of the United States of America": "Proc. Natl. Acad. Sci. USA",
    "Proceedings of the Physical Society, London": "Proc. Phys. Soc. London",
    "Proceedings of the Physical Society, London, Section A": "Proc. Phys. Soc. London, Sec. A",
    "Proceedings of the Physical Society, London, Section B": "Proc. Phys. Soc. London, Sec. B",
    "Proceedings of the Royal Society of London": "Proc. R. Soc. London",
    "Proceedings of the Royal Society of London, Series A: Mathematical and Physical Sciences": "Proc. R. Soc. London, Ser. A",
    "Progress of Theoretical Physics": "Prog. Theor. Phys.",
    "Publications of the Astronomical Society of the Pacific": "Publ. Astron. Soc. Pac.",
    "Quantum Electronics": "Quantum Electron. (UK) or Quantum Electron. (USA)",
    "Quantum Optics": "Quantum Opt.",
    "Radiation Effects": "Radiat. Eff.",
    "Radio Engineering and Electronic Physics (USSR) [translation of Radiotekhnika i Elektronika]": "Radio Eng. Electron. Phys. (USSR)",
    "Radio Engineering and Electronics (USSR) [translation of Radiotekhnika i Elektronika]": "Radio Eng. Electron. (USSR)",
    "Radiotekhnika i Elektronika [Radio Engineering and Electronic Physics (USSR)]": "Radiotekh. Elektron. [Radio Eng. Electron. Phys. (USSR)]",
    "Radiotekhnika i Elektronika [Radio Engineering and Electronics (USSR)]": "Radiotekh. Elektron. [Radio Eng. Electron. (USSR)]",
    "RCA Review": "RCA Rev.",
    "Reports on Progress in Physics": "Rep. Prog. Phys.",
    "Review of Scientific Instruments": "Rev. Sci. Instrum.",
    "Reviews of Modern Physics": "Rev. Mod. Phys.",
    "Revista Mexicana de Astronomica y Astrofisica": "Rev. Mex. Astron. Astrofis.",
    "Revista Mexicana de Fisica": "Rev. Mex. Fis.",
    "Revue d'Optique, Théorique et Instrumentale": "Rev. Opt. Theor. Instrum.",
    "Russian Journal of Physical Chemistry [translation of Zhurnal Fizicheskoi Khimii]": "Russ. J. Phys. Chem.",
    "Science": "Science",
    "Scientific American": "Sci. Am.",
    "Solar Physics": "Sol. Phys.",
    "Solid State Communications": "Solid State Commun.",
    "Solid–State Electronics": "Solid–State Electron.",
    "Soviet Astronomy [translation of Astronomicheskii Zhurnal]": "Sov. Astron.",
    "Soviet Astronomy Letters [translation of Pisma v Astronomicheskii Zhurnal]": "Soc. Astron. Lett.",
    "Soviet Journal of Atomic Energy [translation of Atomnaya Energiya]": "Sov. J. At. Energy",
    "Soviet Journal of Low Temperature Physics [translation of Fizika Nizkikh Temperatur]": "Sov. J. Low Temp. Phys.",
    "Soviet Journal of Nuclear Physics [translation of Yadernaya Fizika]": "Sov. J. Nucl. Phys.",
    "Soviet Journal of Optical Technology [translation of Optiko–Mekhanicheskaya Promyshlennost]": "Sov. J. Opt. Technol.",
    "Soviet Journal of Particles and Nuclei [translation of Fizika Elementarnykh Chastitsi Atomnogo Yadra]": "Sov. J. Part. Nucl.",
    "Soviet Journal of Plasma Physics [translation of Fizika Plazmy]": "Sov. J. Plasma Phys.",
    "Soviet Journal of Quantum Electronics [translation of Kvantovaya Elektronika (Moscow)]": "Sov. J. Quantum Electron.",
    "Soviet Physics–Acoustics [translation of Akusticheskii Zhurnal]": "Sov. Phys. Acoust.",
    "Soviet Physics–Crystallography [translation of Kristallografiya]": "Sov. Phys. Crystallogr.",
    "Soviet Physics–Doklady [translation of Doklady Akademii Nauk SSSR]": "Sov. Phys. Dokl.",
    "Soviet Physics–JETP [translation of Zhurnal Eksperimentalnoi i Teoreticheskoi Fiziki]": "Sov. Phys. JETP",
    "Soviet Physics Journal [translation of Izvestiya Vysshikh Uchebnykh Zavedenii, Fizika]": "Sov. Phys. J.",
    "Soviet Physics–Semiconductors [translation of Fizika i Tekhnika Poluprovodnikov]": "Sov. Phys. Semicond.",
    "Soviet Physics–Solid State [translation of Fizika Tverdogo Tela (Leningrad)]": "Sov. Phys. Solid State",
    "Soviet Physics–Technical Physics [translation of Zhurnal Tekhnicheskoi Fiziki]": "Sov. Phys. Tech. Phys.",
    "Soviet Physics–Uspekhi [translation of Uspekhi Fizicheskikh Nauk]": "Sov. Phys. Usp.",
    "Soviet Radiophysics [translation of Izvestiya Vysshikh Uchebnykh Zavedenii, Radiofizika]": "Sov. Radiophys.",
    "Soviet Technical Physics Letters [translation of Pisma v Zhurnal Tekhnicheskoi Fiziki]": "Sov. Tech. Phys. Lett.",
    "Spectrochimica Acta": "Spectrochim. Acta",
    "Spectrochimica Acta, Part A: Molecular Spectroscopy": "Spectrochim. Acta, Part A",
    "Spectrochimica Acta, Part B: Atomic Spectroscopy": "Spectrochim. Acta, Part B",
    "Surface Science": "Surf. Sci.",
    "Teplofizika Vysokikh Temperatur [High Temperature (USSR)]": "Teplofiz. Vys. Temp. [High Temp. (USSR)]",
    "Theoretica Chimica Acta": "Theor. Chim. Acta",
    "Thin Solid Films": "Thin Solid Films",
    "Transactions of the American Crystallographic Association": "Trans. Am. Crystallogr. Assoc.",
    "Transactions of the American Geophysical Union": "Trans. Am. Geophys. Union",
    "Transactions of the American Institute of Mining, Metallurgical and Petroleum Engineers": "Trans. Am. Inst. Min. Metall. Pet. Eng.",
    "Transactions of the American Nuclear Society": "Trans. Am. Nucl. Soc.",
    "Transactions of the American Society for Metals": "Trans. Am. Soc. Met.",
    "Transactions of the American Society of Mechanical Engineers": "Trans. Am. Soc. Mech. Eng.",
    "Transactions of the British Ceramic Society": "Trans. Br. Ceram. Soc.",
    "Transactions of the Faraday Society": "Trans. Faraday Soc.",
    "Transactions of the Metallurgical Society of AIME": "Trans. Metall. Soc. AIME",
    "Transactions of the Society of Rheology": "Trans. Soc. Rheol.",
    "Ukrainian Physics Journal [translation of Ukrainskii Fizicheskii Zhurnal (Russian Edition)]": "Ukr. Phys. J.",
    "Ultrasonics": "Ultrasonics",
    "Uspekhi Fizicheskikh Nauk [Soviet Physics–Uspekhi]": "Usp. Fiz. Nauk [Sov. Phys. Usp.]",
    "Vistas in Astronomy": "Vistas Astron.",
    "Wuli Xuebao (Acta Physica Sinica) [Chinese Journal of Physics]": "Wuli Xuebao (Acta Phys. Sin.) [Chin. J. Phys.]",
    "Yadernaya Fizika [Soviet Journal of Nuclear Physics]": "Yad. Fiz. [Sov. J. Nucl. Phys.]",
    "Zeitschrift für Analytische": "Chemie Z. Anal. Chem.",
    "Zeitschrift für Angewandte Physik": "Z. Angew. Phys.",
    "Zeitschrift für Anorganische und Allgemeine": "Chemie Z. Anorg. Allg. Chem.",
    "Zeitschrift für Astrophysik": "Z. Astrophys.",
    "Zeitschrift für Elektrochemie": "Z. Elektrochem.",
    "Zeitschrift für Kristallographie, Kristallgeometrie, Kristallphysik, Kristallchemie": "Z. Kristallogr. Kristallgeom. Kristallphys. Kristallchem.",
    "Zeitschrift für Metallkunde": "Z. Metallkd.",
    "Zeitschrift für Naturforschung": "Z. Naturforsch.",
    "Zeitschrift für Naturforschung, Teil A: Physik, Physikalische Chemie, Kosmophysik": "Z. Naturforsch. Teil A",
    "Zeitschrift für Physik": "Z. Phys.",
    "Zeitschrift für Physik A: Atoms and Nuclei": "Z. Phys. A",
    "Zeitschrift für Physik B: Condensed Matter and Quanta": "Z. Phys. B",
    "Zeitschrift für Physik C: Particles and Fields": "Z. Phys. C",
    "Zeitschrift für Physikalisch–Chemische Materialforschung": "Z. Phys. Chem. Materialforsch.",
    "Zeitschrift für Physikalische Chemie, Abteilung A: Chemische Thermodynamik, Kinetik, Elektrochemie, Eigenschaftslehre": "Z. Phys. Chem. Abt. A",
    "Zeitschrift für Physikalische Chemie, Abteilung B: Chemie der Elementarprozesse, Aufbau der Materie": "Z. Phys. Chem. Abt. B",
    "Zeitschrift für Physikalische Chemie (Frankfurt am Main)": "Z. Phys. Chem. (Frankurt am Main)",
    "Zeitschrift für Physikalische Chemie (Leipzig)": "Z. Phys. Chem. (Leipzig)",
    "Zhurnal Eksperimentalnoi i Teoreticheskoi Fiziki [Soviet Physics–JETP]": "Zh. Eksp. Teor. Fiz. [Sov. Phys. JETP]",
    "Zhurnal Fizicheskoi Khimii [Russian Journal of Physical Chemistry]": "Zh. Fiz. Khim. [Russ. J. Phys. Chem.]",
    "Zhurnal Prikladnoi Spektroskopii [Journal of Applied Spectroscopy (USSR)]": "Zh. Prikl. Spektrosk. [J. Appl. Spectrosc. (USSR)]",
    "Zhurnal Tekhnicheskoi Fiziki [Soviet Physics–Technical Physics]": "Zh. Tekh. Fiz. [Sov. Phys. Tech. Phys.]",
}
# Allow for journal titles without parts after first colon
journal_abbreviation_pairs.update(
    {key.split(':')[0]:journal_abbreviation_pairs[key] for key in journal_abbreviation_pairs if ':' in key}
)

journal_full_names = [p[0] for p in journal_abbreviation_pairs]
journal_abbreviations = [p[1] for p in journal_abbreviation_pairs]



"""Search the INSPIRE database

Documentation can be found here: <https://inspirehep.net/info/hep/api>

"""

from functools import lru_cache

api_url = "https://inspirehep.net/search"


@lru_cache()
def query(pattern, output_format='recjson', output_tags=None, records_in_groups_of=None, jump_to_records=None):
    """Search the INSPIRE database

    API documentation: <https://inspirehep.net/info/hep/api>
    Search documentation: <https://inspirehep.net/info/hep/search-tips>

    Parameters
    ----------
    pattern : str
        This is the query in the Inspire search syntax. All search features and
        operators familiar from the Inspire web interface and documented in the
        online help are supported and complex queries are possible.  Documentation
        here: <https://inspirehep.net/info/hep/search-tips>.
    output_format : str, optional [defaults to 'recjson']
        The format of the response sent back to the client. There are two choices,
        'xm' for (MARC-)XML or 'recjson' for JSON. The XML response format is
        MARCXML or fragments thereof when individual fields are selected via the
        output_tags parameter
    output_tags : str, optional [defaults to None]
        If present, this selects (filter) specific tags from the response. If
        output_format is 'xm', this option takes a comma separated list of MARC
        tags; valid MARC tags for Inspire records can be found here
        <https://twiki.cern.ch/twiki/bin/view/Inspire/DevelopmentRecordMarkup>.  If
        output_format is 'recjson', this is similar to selecting MARC tags, however
        by name instead of numerical value.  In addition the JSON response can
        contain derived or dynamically calculated values which are not available in
        MARC.  If this argument is None, the default output will be returned, which
        generally includes all available information for each record.
    records_in_groups_of : int, optional [defaults to None]
        If present, this parameter specifies the number of records per chunk for
        long responses. Note that the default setting is 25 records per chunk. The
        maximum number is 250.
    jump_to_records : int, optional [defaults to None]
        Long responses are split into several chunks. To access subsequent chunks
        specify the record offset with this parameter.

    Returns
    -------
    json : list or dict
        Usually, this will be a list of the results from the query, even if there
        is only one.  Each result will be a dictionary mapping the requested
        'output_tags' to their corresponding values.

    Raises
    ------
    requests.exceptions.HTTPError :
        If the HTTP request to INSPIRE failed for any reason

    """
    import sys
    import requests

    params = {
        'p': pattern,
        'of': output_format,
    }
    if output_tags is not None:
        params['ot'] = output_tags
    if records_in_groups_of is not None:
        params['rg'] = records_in_groups_of
    if jump_to_records is not None:
        params['jrec'] = jump_to_records
    
    r = requests.get(api_url, params=params)
    if r.status_code != 200:
        print('An error occurred when trying to access <{0}>.'.format(api_url), file=sys.stderr)
        try:
            print(r.json(), file=sys.stderr)
        except:
            pass
        r.raise_for_status()
        raise RuntimeError()  # Will only happen if the response was not strictly an error
    return r.json()


def extract_bibtex_key(system_control_numbers):
    """Get bibtex key from 'system_control_numbers' field
    
    Unfortunately, this seems to be the only way to get the bibtex key.  I have
    seen suggestions around the github issues for inspirehep/invenio that this
    should always be present

    """
    if isinstance(system_control_numbers, dict):
        system_control_numbers = [system_control_numbers,]
    bibtex_keys = [number.get('value', '') for number in system_control_numbers
                   if number.get('institute', '') in ['INSPIRETeX', 'SPIRESTeX']]
    bibtex_keys = [key for key in bibtex_keys if key]
    if not bibtex_keys:
        return ''
    return bibtex_keys[0]


def extract_doi(doi):
    """Ensure that 'doi' is a single string
    
    Occasionally, INSPIRE returns a list of identical DOIs.  This just extracts the
    first element if it is such a list, or returns the input otherwise.

    """
    if isinstance(doi, list) and len(doi)>0:
        return doi[0]
    return doi


def extract_doi_url(doi):
    """Ensure that 'doi' is a single string
    
    Occasionally, INSPIRE returns a list of identical DOIs.  This just extracts the
    first element if it is such a list, or returns the input otherwise.

    """
    doi = extract_doi(doi)
    if doi:
        return 'https://dx.doi.org/' + doi
    else:
        return doi


def extract_arxiv_url(system_control_numbers):
    """Extract any arxiv URLs from the system_control_number field

    """
    if isinstance(system_control_numbers, dict):
        system_control_numbers = [system_control_numbers,]
    arxiv_urls = [
        number['value'].replace('oai:arXiv.org:', 'https://arxiv.org/abs/')
        for number in system_control_numbers if number.get('institute', '') == 'arXiv' and 'value' in number
    ]
    if not arxiv_urls:
        return ''
    return arxiv_urls[0]


def map_bibtex_keys_to_doi(bibtex_keys):
    """Map a list of INSPIRE bibtex keys to DOIs

    This function queries the INSPIRE database, searching for each of the input
    bibtex keys, and extracting the corresponding DOIs (if present on INSPIRE).

    Parameter
    ---------
    bibtex_keys : str, or list of str
        Each string should be precisely one bibtex key from INSPIRE.  Note that any
        bibtex keys that are not found by INSPIRE are simply ignored.

    Returns
    -------
    key_to_doi : dict
        The output is a dictionary mapping each input bibtex key itself to the
        corresponding DOI.  These are raw DOIs; to get a URL, just append the DOI
        to 'https://dx.doi.org/', which will resolve to the appropriate web page
        for that DOI.  Note that any record that is found by its bibtex key but
        does not have a DOI will simply be absent from this dictionary.

    Raises
    ------
    requests.exceptions.HTTPError
        If the HTTP request to INSPIRE failed for any reason.

    """
    if not isinstance(bibtex_keys, list):
        bibtex_keys = [bibtex_keys,]
    pattern = 'find texkey ' + ' or texkey '.join(bibtex_keys)
    output_tags = 'system_control_number,doi'
    results = query(pattern, output_tags=output_tags)
    mapping = {
        bibtex_key: doi
        for result in results
        for bibtex_key in [extract_bibtex_key(result['system_control_number']),]
        for doi in [extract_doi(result['doi']),]
        if bibtex_key and doi
    }
    return mapping


def map_bibtex_keys_to_identifiers(bibtex_keys):
    """Map a list of INSPIRE bibtex keys to DOIs, arxiv numbers, or URLs

    This function queries the INSPIRE database, searching for each of the input
    bibtex keys, and extracting the corresponding DOIs (if present on INSPIRE).
    Failing that

    Parameter
    ---------
    bibtex_keys : str, or list of str
        Each string should be precisely one bibtex key from INSPIRE.  Note that any
        bibtex keys that are not found by INSPIRE are simply ignored.

    Returns
    -------
    key_to_identifier : dict
        The output is a dictionary mapping each input bibtex key itself to the
        corresponding DOI URL, arxiv URL, or other URL.

    Raises
    ------
    requests.exceptions.HTTPError
        If the HTTP request to INSPIRE failed for any reason.

    """
    if not bibtex_keys:
        return {}
    if not isinstance(bibtex_keys, list):
        bibtex_keys = [bibtex_keys,]
    pattern = 'find texkey ' + ' or texkey '.join(bibtex_keys)
    output_tags = 'system_control_number,doi'
    results = query(pattern, output_tags=output_tags)
    mapping = {
        bibtex_key: (doi if doi else arxiv)
        for result in results
        for bibtex_key in [extract_bibtex_key(result['system_control_number']),]
        for doi in [extract_doi_url(result['doi']),]
        for arxiv in [extract_arxiv_url(result['system_control_number']),]
        if bibtex_key and (bool(doi) or bool(arxiv))
    }
    return mapping



import pickle
import copy
import numpy as np
import pytest
import sxs


def test_creation_failures():
    np.random.seed(1234)
    n_times = 57

    # Ask to store in Fortran order
    with pytest.raises(ValueError):
        sxs.TimeSeries(np.random.rand(13, 7, 2), np.linspace(0, 10, num=13), time_axis=0, order="F")
    with pytest.raises(ValueError):
        sxs.TimeSeries(np.random.rand(13, 7, 2), np.linspace(0, 10, num=13), time_axis=0, dtype=float, order="F")

    # Pass more than one positional argument
    with pytest.raises(ValueError):
        sxs.TimeSeries(np.random.rand(13, 7, 2), np.linspace(0, 10, num=13), np.random.rand(2))

    # Pass 0-dimensional input array
    with pytest.raises(ValueError):
        sxs.TimeSeries(np.array(3.0), np.linspace(0, 10))
    with pytest.raises(ValueError):
        sxs.TimeSeries(np.array(3.0), np.array(4.0))

    # Pass input array with non-finite values
    for non_finite in [np.inf, -np.inf, np.nan]:
        t = np.linspace(0, n_times)
        a = np.random.rand(n_times, 17)
        a[1, 2] = non_finite
        with pytest.raises(ValueError):
            sxs.TimeSeries(a, t)

    # Forget to specify time array
    with pytest.raises(ValueError):
        sxs.TimeSeries(np.random.rand(n_times, 17))

    # Pass complex time array
    with pytest.raises(ValueError):
        sxs.TimeSeries(np.random.rand(n_times, 17), np.linspace(0, n_times) + 0j)

    # Pass bool time array
    with pytest.raises(ValueError):
        sxs.TimeSeries(np.random.rand(n_times, 17), np.random.randint(0, 2, size=n_times, dtype=bool))

    # Pass time array with ndim != 1
    with pytest.raises(ValueError):
        sxs.TimeSeries(np.random.rand(n_times, 17), np.array(3.0))
    with pytest.raises(ValueError):
        sxs.TimeSeries(np.random.rand(n_times, 17), np.linspace(0, n_times, 18).reshape((-1, 2)))

    # Pass time array with non-finite values
    for non_finite in [np.inf, -np.inf, np.nan]:
        t = np.linspace(0, 10, num=n_times)
        a = np.random.rand(n_times, 17)
        t[3] = non_finite
        with pytest.raises(ValueError):
            sxs.TimeSeries(a, t)

    # Pass time array with non-monotonic values
    t = np.linspace(0, 10, num=n_times)
    a = np.random.rand(n_times, 17)
    t[3] = t[2]
    with pytest.raises(ValueError):
        sxs.TimeSeries(a, t)
    t[3] = (t[1] + t[2]) / 2.0
    with pytest.raises(ValueError):
        sxs.TimeSeries(a, t)

    # Mismatched sizes
    for d_size in [-2, -1, 1, 2]:
        t = np.linspace(0, 10, num=n_times)
        a = np.random.rand(n_times+d_size)
        with pytest.raises(ValueError):
            sxs.TimeSeries(a, t)
        a = np.random.rand(n_times+d_size, 7)
        with pytest.raises(ValueError):
            sxs.TimeSeries(a, t)
        a = np.random.rand(7, n_times+d_size)
        with pytest.raises(ValueError):
            sxs.TimeSeries(a, t)

    # Incorrectly specified time_axis
    t = np.linspace(0, 10, num=n_times)
    a = np.random.rand(n_times, 7)
    with pytest.raises(ValueError):
        sxs.TimeSeries(a, t, time_axis=1)
    a = np.random.rand(7, n_times)
    with pytest.raises(ValueError):
        sxs.TimeSeries(a, t, time_axis=0)

def test_creation_successes():
    np.random.seed(1234)
    n_times = 57

    shapes = [
        (n_times,),
        (n_times, 7),
        (n_times, 3, 7),
        (3, n_times, 7),
        (3, 7, n_times),
    ]
    axes = [
        0,
        0,
        0,
        1,
        2,
    ]
    for shape, axis in zip(shapes, axes):
        t = np.linspace(0, 10, num=n_times)
        a = np.random.rand(*shape)
        b = sxs.TimeSeries(a, t)
        assert b.time_axis == axis
        assert np.array_equal(b.ndarray, a)
        assert b.base is a
        i = (0,)*b.ndim
        b[i] = 1.2
        assert a[i] == 1.2
        b = sxs.TimeSeries(a, time=t)
        assert b.time_axis == axis
        assert np.array_equal(b.ndarray, a)
        b = sxs.TimeSeries(a, time=t, time_axis=axis)
        assert b.time_axis == axis
        assert np.array_equal(b.ndarray, a)


def test_view():
    np.random.seed(1234)
    n_times = 57

    shapes = [
        (n_times,),
        (n_times, 7),
        (n_times, 3, 7),
        (3, n_times, 7),
        (3, 7, n_times),
    ]
    axes = [
        0,
        0,
        0,
        1,
        2,
    ]
    for shape, axis in zip(shapes, axes):
        t = np.linspace(0, 10, num=n_times)
        a = np.random.rand(*shape)
        b = a.view(sxs.TimeSeries)


def test_slice():
    np.random.seed(1234)
    n_times = 57

    shapes = [
        (n_times,),
        (n_times, 7),
        (n_times, 3, 7),
        (3, n_times, 7),
        (3, 7, n_times),
    ]
    axes = [
        0,
        0,
        0,
        1,
        2,
    ]
    take_indices = np.array([0, 2, -1])
    for shape, axis in zip(shapes, axes):
        t = np.linspace(0, 10, num=n_times)
        a = np.random.rand(*shape)
        b = sxs.TimeSeries(a, t)
        # indices = tuple(slice(None) if i != b.time_axis else take_indices for i in range(b.ndim))
        # c = b[indices]
        # assert isinstance(c, type(b))
        # assert c.time_axis == b.time_axis
        # shape = list(b.shape)
        # shape[b.time_axis] = len(take_indices)
        # assert c.shape == tuple(shape)
        # assert c.n_times == len(take_indices)
        # assert np.array_equal(c.time, b.time[take_indices])
        # # c = np.take(b, take_indices, axis=b.time_axis)

    a = np.array([
        [0.0, 1.0, 2.0],
        [-1.0, 0.0, 1.0],
        [-2.0, -1.0, 0.0]
    ])
    b = sxs.TimeSeries(a, time=np.array([-1, 0, 1]), time_axis=0)
    with pytest.raises(ValueError):
        b[b < 0.0]


def test_basic_slicing():
    np.random.seed(1234)
    n_times = 57
    t = np.linspace(0, 10, num=n_times)

    for axis in [0, 1, 2, 3]:
        shape = [7, 9, 11, 12]
        shape[axis] = n_times
        a = np.random.rand(*shape)
        b = sxs.TimeSeries(a, time=t, time_axis=axis)

        # Only key is newaxis
        c = b[np.newaxis]
        assert type(c) == type(b)
        assert c.time_axis == b.time_axis + 1
        assert c.shape[0] == 1
        assert c.shape[1:] == b.shape
        assert np.array_equal(c.ndarray[0], b.ndarray)

        # Only key is a single integer
        c = b[2]
        assert type(c) == type(b)
        if axis == 0:
            assert c.time_axis == b.time_axis
            assert c.shape[1:] == b.shape[1:]
            assert c.shape[0] == 1
            assert np.array_equal(c.ndarray[0], b.ndarray[2])
            assert np.array_equal(c.time, b.time[2:3])
        else:
            assert c.time_axis == b.time_axis - 1
            assert c.shape == b.shape[1:]
            assert np.array_equal(c.ndarray, b.ndarray[2])
            assert np.array_equal(c.time, b.time)

        # Only key is a slice
        c = b[1:4]
        assert type(c) == type(b)
        assert c.time_axis == b.time_axis
        assert c.shape[0] == 3
        assert c.shape[1:] == b.shape[1:]
        assert np.array_equal(c.ndarray, b.ndarray[1:4])
        if axis == 0:
            assert np.array_equal(c.time, b.time[1:4])
        else:
            assert np.array_equal(c.time, b.time)

        # Only key is an Ellipsis
        c = b[...]
        assert type(c) == type(b)
        assert c.time_axis == b.time_axis
        assert c.shape == b.shape
        assert np.array_equal(c.ndarray, b.ndarray)
        assert np.array_equal(c.time, b.time)
        
        # Ellipsis + last key is a single integer
        c = b[..., 2]
        assert type(c) == type(b)
        if axis == b.ndim - 1:
            assert c.time_axis == b.time_axis
            assert c.shape[:-1] == b.shape[:-1]
            assert c.shape[-1] == 1
            assert np.array_equal(c.ndarray[..., 0], b.ndarray[..., 2])
            assert np.array_equal(c.time, b.time[2:3])
        else:
            assert c.time_axis == b.time_axis
            assert c.shape == b.shape[:-1], (axis, a.shape, b.shape, c.shape)
            assert np.array_equal(c.ndarray, b.ndarray[..., 2])
            assert np.array_equal(c.time, b.time)

        # Ellipsis + last key is a slice
        c = b[..., 1:4]
        assert type(c) == type(b)
        assert c.time_axis == b.time_axis
        assert c.shape[:-1] == b.shape[:-1]
        assert c.shape[-1] == 3
        assert np.array_equal(c.ndarray, b.ndarray[..., 1:4])
        if axis == b.ndim - 1:
            assert np.array_equal(c.time, b.time[1:4])
        else:
            assert np.array_equal(c.time, b.time)

        # First key is a single integer + Ellipsis
        c = b[2, ...]
        assert type(c) == type(b)
        if axis == 0:
            assert c.time_axis == b.time_axis
            assert c.shape[1:] == b.shape[1:]
            assert c.shape[0] == 1
            assert np.array_equal(c.ndarray[0], b.ndarray[2])
            assert np.array_equal(c.time, b.time[2:3])
        else:
            assert c.time_axis == b.time_axis - 1
            assert c.shape == b.shape[1:]
            assert np.array_equal(c.ndarray, b.ndarray[2])
            assert np.array_equal(c.time, b.time)

        # First key is a slice + Ellipsis
        c = b[1:4, ...]
        assert type(c) == type(b)
        assert c.time_axis == b.time_axis
        assert c.shape[0] == 3
        assert c.shape[1:] == b.shape[1:]
        assert np.array_equal(c.ndarray, b.ndarray[1:4])
        if axis == 0:
            assert np.array_equal(c.time, b.time[1:4])
        else:
            assert np.array_equal(c.time, b.time)

        # Single int + slice
        c = b[1, 1:4]
        assert type(c) == type(b)
        if axis == 0:
            assert c.shape[0] == 1
            assert c.shape[1] == 3
            assert c.shape[2:] == b.shape[2:]
            assert c.time_axis == b.time_axis
            assert np.array_equal(c.ndarray, b.ndarray[1:2, 1:4])
            assert np.array_equal(c.time, b.time[1:2])
        elif axis == 1:
            assert c.shape[0] == 3
            assert c.shape[1:] == b.shape[2:]
            assert c.time_axis == b.time_axis - 1
            assert np.array_equal(c.ndarray, b.ndarray[1, 1:4])
            assert np.array_equal(c.time, b.time[1:4])
        else:
            assert c.shape[0] == 3
            assert c.shape[1:] == b.shape[2:]
            assert c.time_axis == b.time_axis - 1
            assert np.array_equal(c.ndarray, b.ndarray[1, 1:4])
            assert np.array_equal(c.time, b.time)

        # Slice + single int
        c = b[1:4, 2]
        assert type(c) == type(b)
        if axis == 0:
            assert c.shape[0] == 3
            assert c.shape[1:] == b.shape[2:]
            assert c.time_axis == b.time_axis
            assert np.array_equal(c.ndarray, b.ndarray[1:4, 2])
            assert np.array_equal(c.time, b.time[1:4])
        elif axis == 1:
            assert c.shape[0] == 3
            assert c.shape[1] == 1
            assert c.shape[2:] == b.shape[2:]
            assert c.time_axis == b.time_axis
            assert np.array_equal(c.ndarray, b.ndarray[1:4, 2:3])
            assert np.array_equal(c.time, b.time[2:3])
        else:
            assert c.shape[0] == 3
            assert c.shape[1:] == b.shape[2:]
            assert c.time_axis == b.time_axis - 1
            assert np.array_equal(c.ndarray, b.ndarray[1:4, 2])
            assert np.array_equal(c.time, b.time)

        # Add a new axis in front of int-indexed first dimension
        c = b[np.newaxis, 3]
        assert type(c) == type(b)
        if axis == 0:
            assert c.shape[0] == 1
            assert c.shape[1] == 1
            assert c.shape[2:] == b.shape[1:]
            assert c.time_axis == b.time_axis + 1
            assert np.array_equal(c.ndarray, b.ndarray[np.newaxis, 3:4])
            assert np.array_equal(c.time, b.time[3:4])
        else:
            assert c.shape[0] == 1
            assert c.shape[1:] == b.shape[1:]
            assert c.time_axis == b.time_axis
            assert np.array_equal(c.ndarray, b.ndarray[np.newaxis, 3])
            assert np.array_equal(c.time, b.time)

        # Add a new axis between two int-indexed dimensions
        c = b[2, np.newaxis, 3]
        assert type(c) == type(b)
        if axis == 0:
            assert c.shape[0] == 1
            assert c.shape[1] == 1
            assert c.shape[2:] == b.shape[2:]
            assert c.time_axis == b.time_axis
            assert np.array_equal(c.ndarray, b.ndarray[2:3, np.newaxis, 3])
            assert np.array_equal(c.time, b.time[2:3])
        elif axis == 1:
            assert c.shape[0] == 1
            assert c.shape[1] == 1
            assert c.shape[2:] == b.shape[2:]
            assert c.time_axis == b.time_axis
            assert np.array_equal(c.ndarray, b.ndarray[2, np.newaxis, 3:4])
            assert np.array_equal(c.time, b.time[3:4])
        else:
            assert c.shape[0] == 1
            assert c.shape[1:] == b.shape[2:]
            assert c.time_axis == b.time_axis - 1
            assert np.array_equal(c.ndarray, b.ndarray[2, np.newaxis, 3])
            assert np.array_equal(c.time, b.time)

        with pytest.raises(ValueError):
            b[..., 3, ...]


def test_compare():
    np.random.seed(1234)
    n_times = 57
    a = np.array([
        [0.0, 1.0, 2.0],
        [-1.0, 0.0, 1.0],
        [-2.0, -1.0, 0.0]
    ])
    b = sxs.TimeSeries(a, time=np.array([-1, 0, 1]), time_axis=0)
    c = b < 0.0
    assert np.array_equal(b.time, c.time)
    assert b.time_axis == c.time_axis
    assert np.array_equal(a < 0.0, c)
    with pytest.raises(ValueError):
        b[c]


# def test_repr():
#     np.random.seed(1234)
#     n_times = 57
#     t = np.linspace(0, 10, num=n_times)
#     a = np.random.rand(n_times, 3, 5)
#     b = sxs.TimeSeries(a, time=t, time_axis=0)
#     print('repr:', repr(b))
#     print('str:', str(b))
    



from .references import limit_author_list
from .arxiv import get_entry_by_arxiv_id, get_journal_reference, get_submission_comment


def format_paper_list_entry(arxiv_id, submission_comment='', desired_authors=None, author_list_length_limit=10):
    import re
    from pylatexenc.latexencode import utf8tolatex
    import textwrap
    wrapper = textwrap.TextWrapper(initial_indent='  ', subsequent_indent='    ', width=70,
                                   break_long_words=False, break_on_hyphens=False)
    entry = get_entry_by_arxiv_id(arxiv_id)
    authors = [author['name'] for author in entry['authors']]
    authors = limit_author_list(authors, desired_authors=desired_authors, author_list_length_limit=author_list_length_limit)
    authors = [utf8tolatex(a).replace(r'{\textasciitilde}', '~') for a in authors]
    author_list = ', '.join(authors)
    title = r'\textit{{{0}}}'.format(entry['title'])
    submission_comment = submission_comment.strip()
    if not submission_comment:
        journal_ref = get_journal_reference(entry, r'{pub} \textbf{{{volume}}}, {issue} ({year})')
        if journal_ref:
            submission_comment = ', published as ' + journal_ref.strip()
        else:
            submission_comment = get_submission_comment(entry)
    if submission_comment and not submission_comment.startswith(','):
        submission_comment = ', ' + submission_comment
    arxiv_url = re.sub('v[0-9]*$', '', entry['id'].replace('http://', 'https://'))
    entry = r'\item {0}, {1}{2}; available at \url{{{3}}}'.format(author_list, title, submission_comment, arxiv_url)
    entry = wrapper.fill(' '.join(entry.split()))
    return entry



import numpy as np
import scipy


def align1d(wa, wb, t1, t2, n_brute_force=None):
    """Align waveforms by shifting in time

    This function determines the optimal time offset to apply to `wb` by minimizing
    the averaged (over time) squared difference in the L² norms (over the sphere)
    of the waveforms:
    
        ∫ [ ‖wa(t)‖ - ‖wb(t+δt)‖ ]² dt

    The integral is taken from time `t1` to `t2`.

    No changes are actually made to the input waveforms, but the result of this
    function can be used to offset the second waveform as in

        δt = align1d(wa, wb, t1, t2)
        wb.t -= δt

    As always, be careful, because `wb.t` is a reference to a numpy array that may
    be shared among copies of the data; use `wb.t = wb.t - δt` if you want to
    create a new copy of that array.

    Note that the input waveforms are assumed to be initially aligned at least well
    enough that:
    
      1) the time span from `t1` to `t2` in the two waveforms will overlap at
         least slightly after the second waveform is shifted in time; and
      2) waveform `wb` contains all the times corresponding to `t1` to `t2` in
         waveform `wa`.

    The first of these can usually be assured by simply aligning the peaks prior to
    calling this function:

        wb.t -= wb.max_norm_time() - wa.max_norm_time()

    The second assumption will be satisfied as long as `t1` is not too close to the
    beginning of `wb` and `t2` is not too close to the end.

    Parameters
    ----------
    wa : WaveformModes
    wb : WaveformModes
        Waveforms to be aligned
    t1 : float
    t2 : float
        Beginning and end of integration interval
    n_brute_force : int, optional
        Number of evenly spaced δt values between (t1-t2) and (t2-t1) to sample
        for the initial guess.  By default, this is just the maximum number of
        time steps in the range (t1, t2) in the input waveforms.  If this is
        too small, an incorrect local minimum may be found.

    Notes
    -----
    Choosing the time interval is usually the most difficult choice to make when
    aligning waveforms.  Assuming you want to align during inspiral, the times
    must span sufficiently long that the waveforms' norm (equivalently, orbital
    frequency changes) significantly from `t1` to `t2`.  This means that you
    cannot always rely on a specific number of orbits, for example.  Also note
    that neither number should be too close to the beginning or end of either
    waveform, to provide some "wiggle room".

    Precession generally causes no problems for this function.  In principle,
    eccentricity, center-of-mass offsets, boosts, or other supertranslations could
    cause problems, but this function begins with a brute-force method of finding
    the optimal time offset that will avoid local minima in all but truly
    outrageous situations.  In particular, as long as `t1` and `t2` are separated
    by enough, there should never be a problem.

    """
    from scipy.interpolate import CubicSpline
    from scipy.integrate import trapezoid
    from scipy.optimize import least_squares

    # Check that (t1, t2) makes sense and is actually contained in both waveforms
    if t2<=t1:
        raise ValueError(f"(t1,t2)=({t1}, {t2}) is out of order")
    if wa.t[0] > t1 or wa.t[-1] < t2:
        raise ValueError(f"(t1,t2)=({t1}, {t2}) not contained in wa.t, which spans ({wa.t[0]}, {wa.t[-1]})")
    if wb.t[0] > t1 or wb.t[-1] < t2:
        raise ValueError(f"(t1,t2)=({t1}, {t2}) not contained in wb.t, which spans ({wb.t[0]}, {wb.t[-1]})")

    # Figure out time offsets to try
    δt_lower = max(t1 - t2, wb.t[0]-t1)
    δt_upper = min(t2 - t1, wb.t[-1]-t2)

    # We'll start by brute forcing, sampling time offsets evenly at as many
    # points as there are time steps in (t1,t2) in the input waveforms
    if n_brute_force is None:
        n_brute_force = max(sum((wa.t>=t1)&(wa.t<=t2)), sum((wb.t>=t1)&(wb.t<=t2)))
    δt_brute_force = np.linspace(δt_lower, δt_upper, num=n_brute_force)

    # Times at which the differences will be evaluated
    t_reference = wa.t[(wa.t>=t1)&(wa.t<=t2)]

    # Define the cost function
    norm_a = wa.norm.ndarray[(wa.t>=t1)&(wa.t<=t2)]
    norm_b = CubicSpline(wb.t, wb.norm.ndarray)
    normalization = trapezoid((norm_a)**2, t_reference)
    def cost(δt):
        # Take the sqrt because least_squares squares the inputs...
        return np.sqrt(trapezoid((norm_a - norm_b(t_reference+δt))**2, t_reference) / normalization)

    # Optimize by brute force
    cost_brute_force = [cost(δt) for δt in δt_brute_force]
    δt = δt_brute_force[np.argmin(cost_brute_force)]

    # Optimize explicitly
    optimum = least_squares(cost, δt, bounds=(δt_lower, δt_upper))

    return optimum.x[0]



"""Functions for syncing the SimAnnex to Zenodo

"""

def sync(annex_dir='./', lock_file_path='~/.sxs/zenodo_simannex_sync.lock', verbosity=0,
         directories_and_permissions=[['Public', 'open'], ['Private', 'closed']]):
    """Sync the SimAnnex with zenodo

    This function searches for local SXS simulations, then loops through each and compares to what
    is on zenodo, ensuring that they are the same.  Note that exceptions may occur for various
    reasons; if this happens, the exception is intercepted, and the loop over systems continues.

    Parameters
    ----------
    annex_dir : str, optional [defaults to './']
        Absolute or relative path to base directory to search for SXS systems.
    lock_file_path : str, optional [defaults to '~/.sxs/zenodo_simannex_sync.lock']
        Absolute or relative path to file to use as lock file to ensure that only one instance of
        this function runs at a time.  If this is None, no lock file will be used, which means that
        it would be possible to run multiple instances.
    verbosity : int, optional [defaults to 0]
        Amount of output to pass.  If 0, both stdout and stderr will be intercepted; if 1, only
        stdout will be intercepted; if 2 or greater, nothing will be intercepted.  Both will be
        printed if an exception is raised, unless this value is less than 0, in which case nothing
        will be printed even if there is an exception.
    directories_and_permissions : list of pairs, optional
        This should be a list of directories relative to `annex_dir` to be searched for SXS systems,
        along with the access right to use in zenodo for every system found.  The default value is
        [['Public', 'open'], ['Private', 'closed']].

    """
    import contextlib
    import sys
    import io
    import time
    import os.path
    import traceback
    import tqdm
    from ..utilities import lock_file_manager, find_simulation_directories
    from . import upload

    annex_dir = os.path.abspath(os.path.expanduser(annex_dir))

    if lock_file_path is None:
        lfm = contextlib.nullcontext()
    else:
        lfm = lock_file_manager(os.path.expanduser(lock_file_path))

    if verbosity < 1:
        def verbosity_manager(new_target):
            stack = contextlib.ExitStack()
            stack.enter_context(contextlib.redirect_stdout(new_target))
            stack.enter_context(contextlib.redirect_stderr(new_target))
            return stack
        def progress(iterable):
            return iterable
    elif verbosity < 2:
        def verbosity_manager(new_target):
            stack = contextlib.ExitStack()
            stack.enter_context(contextlib.redirect_stdout(new_target))
            return stack
        def progress(iterable):
            return tqdm.tqdm(iterable, dynamic_ncols=True)
    else:
        verbosity_manager = contextlib.nullcontext
        def progress(iterable):
            return tqdm.tqdm(iterable, dynamic_ncols=True)

    with verbosity_manager(None):
        print('Searching for simulations')

    with lfm:
        for root, access_right in directories_and_permissions:
            simulation_directories = find_simulation_directories(os.path.join(annex_dir, root))
            for simulation_directory in progress(simulation_directories):
                time.sleep(0.1)  # Allow for possible keyboard interrupt from this whole loop
                verbose_string = io.StringIO()
                try:
                     with verbosity_manager(verbose_string):
                        upload(simulation_directory, access_right=access_right, skip_checksums='if_file_is_older',
                               skip_existing=False, error_on_existing=False, publish="if_pending")
                        print('\n')
                except KeyboardInterrupt:
                    if verbosity < 0:
                        continue
                    if verbosity < 1:
                        print(verbose_string.getvalue())
                    print("Interrupted in", simulation_directory)
                    print('\n')
                    continue
                except Exception as e:
                    if verbosity < 0:
                        continue
                    if verbosity < 1:
                        print(verbose_string.getvalue())
                    print("Failed in", simulation_directory)
                    print(e)
                    traceback.print_exc()
                    print('\n')



import numpy as np
import quaternionic
import spherical
import pytest
import sxs

from .conftest import constant_waveform, linear_waveform, random_waveform


def test_modes_rotate(h, eps):
    import time

    ϵ = 5 * (2 * h.ell_max + 1) * 2 * eps

    print()
    for i, R in enumerate([quaternionic.one, quaternionic.one * np.ones_like(h.t)]):
        t1 = time.perf_counter()
        hprm = h.rotate(R)
        t2 = time.perf_counter()
        print(f"\tRotation {i+1} took {t2-t1:.4f} seconds")
        assert type(h) == type(hprm)
        assert np.array_equal(h.t, hprm.t)
        assert np.allclose(h.ndarray, hprm.ndarray, rtol=ϵ, atol=ϵ)

        metadata = h._metadata.copy()
        metadataprm = hprm._metadata.copy()
        for d in [metadata, metadataprm]:
            for key in ['time', 'frame']:
                d.pop(key, None)
            for key in ['space_translation', 'boost_velocity']:
                d[key] = d[key].tolist()
        assert metadata == metadataprm


def test_modes_rotate_evaluate(h, Rs, eps):
    """Test that evaluating modes at rotated points == evaluating rotated modes at points"""
    import time

    ell_max = h.ell_max
    ϵ = (2 * h.ell_max + 1) * 2 * eps

    equiangular_grid = spherical.theta_phi(2 * ell_max + 1, 2 * ell_max + 1)
    Rθϕ = quaternionic.array.from_spherical_coordinates(equiangular_grid)

    for i, R in enumerate(Rs):
        hprm = h.copy().rotate(R)  # hprm = h @ 𝔇(R)
        m1 = hprm.evaluate(Rθϕ)  # m1 = hprm @ 𝔇(Rθϕ) √...
        m2 = h.evaluate(R * Rθϕ)  # m2 = h @ 𝔇(R * Rθϕ) √...
        assert np.allclose(m1, m2, rtol=ϵ, atol=ϵ)


@pytest.mark.parametrize("w", [linear_waveform, constant_waveform])
def test_dpa_simple_cases(w, eps):
    LL = w().dominant_eigenvector_LL()
    LL_expected = np.zeros_like(LL, dtype=float)
    LL_expected[:, 2] = 1.0
    assert np.allclose(LL, LL_expected, rtol=eps, atol=eps)


@pytest.mark.parametrize("w", [linear_waveform, constant_waveform])
def test_dpa_rotated_simple_cases(w, Rs):
    # We use `begin=1.0` because we need to avoid situations where the modes
    # are all zeros, which can happen in `linear_waveform` at t=0.0
    W = w(begin=1.0, ell_min=0, n_times=len(Rs))
    LL = W.rotate(Rs.conjugate()).dominant_eigenvector_LL()
    LL_expected = (Rs * quaternionic.array([quaternionic.z for _ in range(len(Rs))]) * Rs.conjugate()).vector

    # with np.printoptions(precision=4, linewidth=180, suppress=True):
    #     print("LL:")
    #     print(LL)
    #     print()
    #     print("LL_expected:")
    #     print(LL_expected)

    # Because the dpa is only defined up to a sign, all we need is for the
    # dot product between the dpa and the expected value to be close to
    # either 1 or -1.  This finds the largest difference, based on the
    # smaller of the two sign options.
    assert (
        max(
            np.amin(
                np.vstack((np.linalg.norm(LL - LL_expected, axis=1), np.linalg.norm(LL + LL_expected, axis=1))), axis=0
            )
        )
        < 1.0e-14
    )


@pytest.mark.parametrize("w", [linear_waveform, constant_waveform, random_waveform])
def test_dpa_rotated_generally(w, Rs):
    np.random.seed(1234)
    n_copies = 10
    W = w(begin=1.0, end=100.0, n_times=n_copies * len(Rs), ell_min=0, ell_max=8)
    R_basis = quaternionic.array([R for R in Rs for _ in range(n_copies)])

    # We use `begin=1.0` because we need to avoid situations where the modes
    # are all zeros, which can happen in `linear_waveform` at t=0.0
    LL1 = (
        R_basis
        * quaternionic.array.from_vector_part(W.dominant_eigenvector_LL())
        * R_basis.conjugate()
    ).vector
    LL2 = W.rotate(R_basis.conjugate()).dominant_eigenvector_LL()

    # if (max(np.amin(np.vstack((np.linalg.norm(LL1 - LL2, axis=1), np.linalg.norm(LL1 + LL2, axis=1))), axis=0))
    #     > 1.0e-12):
    #     with np.printoptions(precision=8, linewidth=180, suppress=True):
    #         print()
    #         print(f"LL1, LL2, LL1-LL2 {LL1.shape}")
    #         print(np.stack((LL1, LL2, LL1-LL2), axis=2))
    #         print()

    # Because the dpa is only defined up to a sign, all we need is for the
    # dot product between the dpa and the expected value to be close to
    # either 1 or -1.  This finds the largest difference, based on the
    # smaller of the two sign options.
    assert (
        max(np.amin(np.vstack((np.linalg.norm(LL1 - LL2, axis=1), np.linalg.norm(LL1 + LL2, axis=1))), axis=0))
        < 1.0e-12
    )


def test_zero_angular_velocity():
    w = constant_waveform(end=10.0, n_times=10000)
    ω = w.angular_velocity
    assert np.allclose(ω, np.zeros_like(ω), atol=1e-15, rtol=0.0)


def test_z_angular_velocity():
    w = constant_waveform(end=10.0, n_times=10000)
    ω = 2 * np.pi / 5.0
    R = np.exp(quaternionic.array.from_vector_part([0, 0, ω / 2]) * w.t)
    w = w.rotate(~R)
    ω_out = w.angular_velocity
    ω_in = np.zeros_like(ω_out)
    ω_in[:, 2] = ω
    assert np.allclose(ω_in, ω_out, atol=1e-12, rtol=2e-8), (
        f"\nω_in = np.array({ω_in.tolist()})\n"
        f"\nω_out = np.array({ω_out.tolist()})\n"
    )


def test_rotated_angular_velocity():
    w = constant_waveform(end=10.0, n_times=10000)
    ω = 2 * np.pi / 5.0
    R0 = quaternionic.array(1, 2, 3, 4).normalized
    R = R0 * np.exp(quaternionic.array.from_vector_part([0, 0, ω / 2]) * w.t)
    w = w.rotate(~R)
    ω = R0 * quaternionic.array.from_vector_part([0, 0, ω]) * R0.inverse
    ω_out = w.angular_velocity
    ω_in = np.zeros_like(ω_out)
    ω_in[:, 0] = ω.x
    ω_in[:, 1] = ω.y
    ω_in[:, 2] = ω.z
    assert np.allclose(ω_in, ω_out, atol=1e-12, rtol=2e-8), (
        f"\nω_in = np.array({ω_in.tolist()})\n"
        f"\nω_out = np.array({ω_out.tolist()})\n"
    )


def test_corotating_frame():
    w = constant_waveform(end=10.0, n_times=100_000)  # Need lots of time steps for accurate integration
    omega = 2 * np.pi / 5.0
    R0 = quaternionic.array.random().normalized
    R_in = R0 * np.exp(quaternionic.array([0, 0, 0, omega / 2]) * w.t)
    w_rot = w.copy().rotate(R_in.conjugate())
    R_out = w_rot.corotating_frame(R0=R0, tolerance=1e-12)
    assert np.allclose(R_in.ndarray, R_out.ndarray, atol=1e-10, rtol=0.0), (
        f"\nR_in = {R_in}\n"
        f"\nR_in-R_out = {R_in-R_out}\n"
        f"\nmax(abs(diff)) = {np.max(np.abs((R_in-R_out).ndarray))}\n"
    )
    w_inertial = w_rot.to_corotating_frame(R0=R0, tolerance=1e-12)
    # with np.printoptions(precision=8, linewidth=180, suppress=True):
    #     print()
    #     print(f"R0 = {R0}")
    #     print(f"R_in = {R_in}")
    #     print(f"R_out = {R_out}")
    #     print(f"w.data = {w.data}")
    #     print(f"w_rot.data = {w_rot.data}")
    #     print(f"w.data-w_rot.data = {w.data-w_rot.data}")
    #     #print(f" = {}")
    assert np.array_equal(w.t, w_inertial.t)
    assert np.allclose(w.data, w_inertial.data, atol=1e-8)
    assert w_inertial.frame_type == "corotating"



import numpy as np


class TimeSeries(np.ndarray):
    # noinspection PyUnresolvedReferences
    """Array-like object representing time-series data

    This object wraps the basic numpy array object, but stores (at least a
    reference to) a corresponding array of time values, and provides several member
    functions for interpolating, differentiating, and integrating.

    Parameters
    ----------
    input_array : (..., N, ...) array_like
        Input data representing the dependent variable, in any form that can be
        converted to a numpy array.  This includes scalars, lists, lists of tuples,
        tuples, tuples of tuples, tuples of lists, and numpy ndarrays.  It can have
        an arbitrary number of dimensions, but the length `N` along `time_axis`
        (see below) must match the length of `time`.  Values must be finite.
    time : (N,) array_like
        1-D array containing values of the independent variable.  Values must be
        real, finite, and in strictly increasing order.
    time_axis : int, optional
        Axis along which `input_array` is assumed to be varying in time, meaning
        that for `time[i]` the corresponding values are `np.take(input_array, i,
        axis=time_axis)`.  If this is not given, the first axis of `input_array`
        that has the same length as `time` is chosen as the time axis — which may
        be prone to errors.

    """

    def __new__(cls, input_array, *args, **kwargs):
        import copy

        dtype = kwargs.pop("dtype", None)
        order = kwargs.pop("order", "C")
        if order == "F":
            raise ValueError(f"Requested array order '{order}' is not supported; it must be 'C'.")
        if len(args) > 1:
            raise ValueError("Only one positional argument may be passed")
        elif len(args) == 1:
            kwargs["time"] = args[0]
        metadata = copy.copy(getattr(input_array, "_metadata", {}))
        metadata.update(**kwargs)

        # Interpret input_array as some type of array
        input_array = np.asanyarray(input_array, dtype=dtype, order=order)
        if input_array.ndim == 0:
            raise ValueError("Input array has 0 dimensions; it must have at least one")
        if not np.all(np.isfinite(input_array)):
            raise ValueError("Input array must contain only finite values.")

        # Get time array
        time = metadata.get("time", None)
        if time is None:
            raise ValueError(
                "Time data must be specified as part of input TimeSeries, as "
                "the second positional argument, or as a keyword argument."
            )
        time = np.asarray(time)
        if np.issubdtype(time.dtype, np.complexfloating):
            raise ValueError("Input `time` must contain real values; it has complex type.")
        if not np.issubdtype(time.dtype, np.number):
            raise ValueError("Input `time` must contain numbers; its dtype is '{time.dtype}'.")
        time = time.astype(float)
        if time.ndim != 1:
            raise ValueError(f"Input `time` array must have exactly 1 dimension; it has {time.ndim}.")
        if not np.all(np.isfinite(time)):
            raise ValueError("Input `time` must contain only finite values.")
        if np.any(np.diff(time) <= 0):
            raise ValueError("Input `time` must be strictly increasing sequence.")
        metadata["time"] = time

        # Get time_axis
        time_axis = metadata.get("time_axis", None)
        if time_axis is None:
            for i in range(input_array.ndim):
                if input_array.shape[i] == time.size:
                    time_axis = i
                    break
        if time_axis is None:
            raise ValueError(
                f"Cannot find axis of size time.size={time.size} in input_array, "
                f"which has shape input_array.shape={input_array.shape}"
            )
        time_axis = time_axis % input_array.ndim  # Map time_axis into [0, input_array.ndim)
        if input_array.shape[time_axis] != time.shape[0]:
            raise ValueError(
                f"input_array.shape[time_axis]={input_array.shape[time_axis]} (with time_axis={time_axis}) "
                f"does not match time.shape[0]={time.shape[0]}"
            )
        metadata["time_axis"] = time_axis

        obj = input_array.view(cls)
        obj._metadata = metadata
        return obj

    def __array_finalize__(self, obj):
        import copy
        if obj is None:
            return
        # # Since ndarray.__array_finalize__ is None, we skip this in its direct descendents:
        # super().__array_finalize__(obj)
        self._metadata = copy.copy(getattr(self, '_metadata', getattr(obj, '_metadata', {})))
        if "time" not in self._metadata:
            self._metadata["time"] = None  # Placeholder about to be updated

    def _slice(self, key):
        """Slice this object

        This is the core function used by __getitem__, and returns not only the sliced
        result, but also the indices used to slice the time array.  This is useful for
        subclasses when they also need the latter to slice additional time-related
        quantities.

        """
        from numbers import Integral

        if isinstance(key, tuple) and len(key) == 0:
            raise ValueError(f"Empty index to {type(self).__name__} does not make sense")

        def newaxis_type(e):
            return isinstance(e, (type(np.newaxis), type(None)))

        def basic_slicing(key_basic):
            """Test if basic slicing occurs given this key_basic

            Numpy's indexing documentation says "Basic slicing occurs when `obj` is
            a `slice` object (constructed by `start:stop:step` notation inside of
            brackets), an integer, or a tuple of slice objects and
            integers. Ellipsis and newaxis objects can be interspersed with these
            as well."

            <https://numpy.org/doc/stable/reference/arrays.indexing.html#basic-slicing-and-indexing>

            """
            if isinstance(key_basic, (Integral, slice, type(Ellipsis))):
                return True
            if not isinstance(key_basic, tuple):
                return False
            for e in key_basic:
                if not isinstance(e, (Integral, slice, type(Ellipsis), type(np.newaxis), type(None))):
                    return False
            return True

        def normalize_basic_slicing_key(key_normalize, ndim):
            """Translate key_basic into full-size tuples, without Ellipsis

            Returns a pair of keys, one appropriate for the original array (with
            `newaxis` removed), and one for the new array (with `newaxis` replaced
            by `slice(None)`).

            """
            if not isinstance(key_normalize, tuple):
                key_normalize = (key_normalize,)
            n_missing = ndim - len(tuple(e for e in key_normalize if not newaxis_type(e)))
            count = key_normalize.count(Ellipsis)
            if count == 0:
                full_key_normalize = key_normalize + (slice(None),) * n_missing
            elif count == 1:
                i_ellipsis = key_normalize.index(Ellipsis)
                full_key_normalize = (
                        key_normalize[:i_ellipsis]
                        + (slice(None),) * (n_missing + 1)
                        + key_normalize[i_ellipsis + 1:]
                )
            else:
                raise ValueError(f"Cannot use more than one Ellipsis in key_basic; found {count}.")
            old_key_normalize = tuple(e for e in full_key_normalize if not newaxis_type(e))
            return old_key_normalize, full_key_normalize

        time_key = slice(None)

        if key is np.newaxis:
            new_time_axis = self.time_axis + 1
            new_data = super().__getitem__(key)

        elif basic_slicing(key):
            old_key, full_key = normalize_basic_slicing_key(key, self.ndim)
            # Find the new time_axis
            i_old, integral_key_correction = 0, 0
            full_time_axis = 0
            for i in range(len(full_key)):
                full_time_axis = i
                key_i = full_key[i]
                if not newaxis_type(key_i):
                    i_old += 1
                if i_old > self.time_axis:
                    break
                if isinstance(key_i, Integral):
                    integral_key_correction += 1
            new_time_axis = full_time_axis - integral_key_correction
            # Expand the dimensions along the time axis if we're taking just one element
            time_key = old_key[self.time_axis]
            if isinstance(time_key, Integral):
                if time_key == -1:
                    time_key = slice(time_key, None)
                else:
                    time_key = slice(time_key, time_key+1)
                full_key = tuple(e if i != full_time_axis else time_key for i, e in enumerate(full_key))
            # Index the data and time arrays
            new_data = super().__getitem__(full_key)

        elif isinstance(key, np.ndarray) and key.ndim-1 <= self.time_axis:
            # This is okay; it's just indexing the leading dimensions
            new_data = super().__getitem__(key)
            new_time_axis = self.time_axis

        else:  # Try the slow way
            new_data = super().__getitem__(key)
            if new_data.ndim != self.ndim:
                raise ValueError(
                    f"\nAdvanced indexing of this {type(self).__name__} object with the key_basic\n\n"
                    + "    " + "\n    ".join(str(key).split("\n")) + "\n\n" +
                    f"changes the shape of the data from {self.shape} to {new_data.shape}.  It is not clear how this\n"
                    f"should affect the time data.  If you still want to index the data like this, extract the\n"
                    f"underlying data as `a.ndarray` and the time as `a.time`, slice them as desired, and reassemble\n"
                    f"them as\n\n"
                    f"    {type(self).__name__}(sliced_data, time=sliced_time, time_axis=sliced_time_axis)\n"
                )
            try:
                time_key = None
                new_time_slice = tuple(0 if i != self.time_axis else slice(None) for i in range(self.ndim))
                new_time = self.time_broadcast[key][new_time_slice].copy()
            except Exception as e:
                raise ValueError(
                    f"\nAdvanced indexing fails when trying to slice time information from this {type(self).__name__}\n"
                    f"object with the key_basic\n\n"
                    + "    " + "\n    ".join(str(key).split("\n")) + "\n\n" +
                    f"The shape of the {type(self).__name__} is {self.shape} and the time axis is {self.time_axis}."
                ) from e
            new_time_axis = self.time_axis

        # Create the new sliced object along with its metadata
        if time_key is not None:
            new_time = self.time[time_key]
        metadata = self._metadata.copy()
        metadata.update(**getattr(new_data, '_metadata', {}))
        metadata["time"] = new_time
        metadata["time_axis"] = new_time_axis

        return type(self)(new_data, **metadata), time_key

    def __getitem__(self, key):
        """Extract a slice of this object

        Note that slicing this object works slightly differently than slicing the
        underlying ndarray object, basically because we want to ensure that the
        returned object is still a TimeSeries object.

        First, if a single element is requested along the time dimension, that
        dimension will not be removed.  For a 2-d ndarray `arr`, taking `arr[3]` will
        return a 1-d array; the first dimension will be removed because only the third
        element is extracted.  For a 2-d TimeSeries `ts` with `time_axis=0`, `ts[3]`
        will return a 2-d TimeSeries; the first dimension will just have size 1,
        representing the third element.  If the requested element is not along the time
        dimension, the requested dimension will be removed as usual.

        Also, taking an irregular slice of this object is not permitted.  For example:

            >>> a = np.arange(3*4).reshape(3, 4)
            >>> a[a % 5 == 0]
            array([ 0,  5, 10])

        Even though `a % 5 == 0` is a 2-d array, indexing flattens `a` and the indexing
        set, so that the result is a 1-d array.  This probably does not make sense for
        TimeSeries arrays, so attempting to do something like this raises a ValueError.

        """
        return self._slice(key)[0]

    @property
    def ndarray(self):
        """View this array as a numpy ndarray"""
        return self.view(np.ndarray)

    @property
    def time_axis(self):
        """Axis of the array along which time varies

        At the time `time[i]`, the corresponding values of the data are
        `np.take(input_array, i, axis=time_axis)`.

        """
        return self._metadata["time_axis"]

    @property
    def time(self):
        """Array of the time steps corresponding to the data"""
        return self._metadata["time"]

    @time.setter
    def time(self, tprime):
        tprime = np.asarray(tprime)
        if np.issubdtype(tprime.dtype, np.complexfloating):
            raise ValueError("Input `tprime` must contain real values; it has complex type.")
        if not np.issubdtype(tprime.dtype, np.number):
            raise ValueError("Input `tprime` must contain numbers; its dtype is '{tprime.dtype}'.")
        tprime = tprime.astype(float)
        if tprime.ndim != 1:
            raise ValueError(f"Input `tprime` array must have exactly 1 dimension; it has {tprime.ndim}.")
        if not np.all(np.isfinite(tprime)):
            raise ValueError("Input `tprime` must contain only finite values.")
        if np.any(np.diff(tprime) <= 0):
            raise ValueError("Input `tprime` must be strictly increasing sequence.")
        if self._metadata["time"] is not None:
            if tprime.shape != self._metadata["time"].shape:
                raise ValueError(
                    f"Input `tprime` must have same shape as original; they are {tprime.shape} "
                    f"and {self._metadata['time'].shape}"
                )
        self._metadata["time"] = tprime

    t = time

    def register_modification(self, func, **kwargs):
        """Add a record of a modification to the metadata

        Note that this function does not actually run the modification; it simply
        records the function name and arguments in this object's metadata.  You are
        expected to run the function for yourself, with the given keyword arguments.

        Also note that the modifications will most likely be written to JSON, so you
        should adjust them to be in basic formats suitable for JSON.  For example, if
        an argument `arr` is ordinarily passed as a numpy array, you should convert to
        a list, with something like `arr.tolist()`.

        Parameters
        ----------
        func : named function
            The function that will modify (or already has modified) this object.  The
            function must have a `__name__` attribute, as most functions do.

        Because we cannot know whether `func` modifies `self` in place or not, we
        cannot design this function to modify the desired result, which is why you must
        call `func` yourself.

        """
        if "modifications" not in self._metadata:
            self._metadata["modifications"] = {}
        self._metadata["modifications"][func.__name__] = kwargs
        return self

    def index_closest_to(self, t):
        """Time index closest to the given time `t`

        Parameters
        ----------
        t : float

        Returns
        -------
        idx : int
            Index such that abs(self.time[idx]-t) is as small as possible

        """
        idx = min(np.searchsorted(self.t, t), self.n_times-1)
        if idx > 0:
            if abs(self.t[idx]-t) > abs(self.t[idx-1]-t):
                idx -= 1
        return idx

    @property
    def n_times(self):
        """Size of the array along the time_axis"""
        return self.time.size

    @property
    def time_broadcast(self):
        """Array of the time steps broadcast to same shape as data

        This property returns a new view (usually involving no copying of memory)
        of the `time` array, with additional dimensions to match the shape of the
        data.

        """
        new_shape = tuple(np.newaxis if i != self.time_axis else slice(None) for i in range(self.ndim))
        # This method is from the comment in np.broadcast_arrays, used so that
        # we don't have to disturb the data array (e.g., copying because of
        # weird ordering).
        # noinspection SpellCheckingInspection
        nditer = np.nditer(
            self.time[new_shape],
            flags=["multi_index", "zerosize_ok"],
            itershape=self.shape,
            order="C"
        )
        return nditer.itviews[0]

    def __repr__(self):
        r = repr(self.ndarray)
        return f"{type(self).__name__}{r[max(0, r.find('(')):-1]}, time={self.time!r}, time_axis={self.time_axis})"

    def __str__(self):
        return repr(self)

    def interpolate(self, new_time, derivative_order=0, out=None):
        """Interpolate this object to a new set of times

        Parameters
        ----------
        new_time : array_like
            Points to evaluate the interpolant at
        derivative_order : int, optional
            Order of derivative to evaluate.  If negative, the antiderivative is
            returned.  Default value of 0 returns the interpolated data without
            derivatives or antiderivatives.  Must be between -3 and 3, inclusive.

        See Also
        --------
        scipy.interpolate.CubicSpline :
            The function that this function is based on.
        antiderivative :
            Calls this funtion with `new_time=self.time` and
            `derivative_order=-antiderivative_order` (defaulting to a single
            antiderivative).
        derivative :
            Calls this function `new_time=self.time` and
            `derivative_order=derivative_order` (defaulting to a single derivative).
        dot :
            Property calling `self.derivative(1)`.
        ddot :
            Property calling `self.derivative(2)`.
        int :
            Property calling `self.antiderivative(1)`.
        iint :
            Property calling `self.antiderivative(2)`.

        Notes
        -----
        This function is essentially a wrapper around `scipy.interpolate.CubicSpline`

        """
        from scipy.interpolate import CubicSpline
        if derivative_order > 3:
            raise ValueError(
                f"{type(self).__name__} interpolation uses CubicSpline, which cannot take a derivative "
                f"of order {derivative_order}."
            )
        new_time = np.asarray(new_time)
        if new_time.ndim != 1:
            raise ValueError(f"New time array must have exactly 1 dimension; it has {new_time.ndim}.")
        new_shape = list(self.shape)
        new_shape[self.time_axis] = new_time.size
        if out is not None:
            out = np.asarray(out)
            if out.shape != new_shape:
                raise ValueError(
                    f"Output array should have shape {new_shape} for consistency with new time array and modes array"
                )
            if out.dtype != self.dtype:
                raise ValueError(
                    f"Output array should have same dtype as this array {self.dtype}; it has dtype {out.dtype}"
                )
        result = out or np.empty(new_shape, dtype=self.dtype)
        spline = CubicSpline(self.time, self.ndarray, axis=self.time_axis)
        if derivative_order < 0:
            spline = spline.antiderivative(-derivative_order)
        elif 0 < derivative_order <= 3:
            spline = spline.derivative(derivative_order)
        result[:] = spline(new_time)
        metadata = self._metadata.copy()
        metadata["time"] = new_time
        metadata["time_axis"] = self.time_axis
        return type(self)(result, **metadata)

    def antiderivative(self, antiderivative_order=1):
        """Integrate modes with respect to time

        Parameters
        ----------
        antiderivative_order : int, optional
            Order of antiderivative to evaluate.  Default value is 1.  Must be between
            -3 and 3, inclusive.

        See Also
        --------
        scipy.interpolate.CubicSpline :
            The function that this function is based on.
        interpolate :
            This function simply calls `self.interpolate` with appropriate arguments.
        int :
            Property calling `self.antiderivative(1)`.
        iint :
            Property calling `self.antiderivative(2)`.

        """
        return self.interpolate(self.time, derivative_order=-antiderivative_order)

    def derivative(self, derivative_order=1):
        """Differentiate modes with respect to time

        Parameters
        ----------
        derivative_order : int, optional
            Order of derivative to evaluate.  Default value is 1.  Must be between -3
            and 3, inclusive.

        See Also
        --------
        scipy.interpolate.CubicSpline :
            The function that this function is based on.
        interpolate :
            This function simply calls `self.interpolate` with appropriate arguments.
        dot :
            Property returning `self.derivative(1)`.
        ddot :
            Property returning `self.derivative(2)`.

        """
        return self.interpolate(self.time, derivative_order=derivative_order)

    @property
    def dot(self):
        """Differentiate modes once with respect to time

        See Also
        --------
        derivative : This property simply returns `self.derivative(1)`
        ddot : Property returning `self.derivative(2)`.

        """
        return self.derivative()

    @property
    def ddot(self):
        """Differentiate modes twice with respect to time

        See Also
        --------
        derivative : This property simply returns `self.derivative(2)`
        dot : Property returning `self.derivative(1)`.

        """
        return self.derivative(2)

    @property
    def int(self):
        """Integrate modes once with respect to time

        See Also
        --------
        antiderivative : This property simply returns `self.antiderivative(1)`
        iint : Property returning `self.antiderivative(2)`.

        """
        return self.antiderivative()

    @property
    def iint(self):
        """Integrate modes twice with respect to time

        See Also
        --------
        antiderivative : This property simply returns `self.antiderivative(2)`
        int : Property returning `self.antiderivative(1)`.

        """
        return self.antiderivative(2)

    def xor(self, reverse=False, preserve_dtype=False, **kwargs):
        """Progressively XOR data along the time axis

        This function steps through an array, starting with the second element, and
        evaluates bitwise XOR on that element and the preceding one.  This is a useful
        step in compressing reasonably continuous data.

        See the documentation of `sxs.utilities.xor` for a full description of this
        function.  Note that this version sets the `axis` argument automatically to be
        the `time_axis`.

        """
        from .utilities import xor
        kw = kwargs.copy()
        kw.update({"axis": self.time_axis})
        return xor(self.ndarray, reverse=reverse, preserve_dtype=preserve_dtype, **kw)

    def truncate(self, abs_tolerance):
        """Truncate the precision of this object's `data` in place

        This function sets bits in the array data to 0 when they have lower
        significance than the number given as or returned by `abs_tolerance`.  This is
        a useful step in compressing data — though it is obviously lossy.

        Parameters
        ----------
        abs_tolerance : {callable, float, array-like}
            If callable, it is called with this object as the parameter, and the
            returned value is treated as a float or array-like would be.  Floats are
            simply treated as a uniform absolute tolerance to be applied at all times.
            Array-like objects must broadcast against this array, and each element is
            treated as the absolute tolerance for all the elements it broadcasts to.

        Returns
        -------
        None
            This value is returned to serve as a reminder that this function operates
            in place.

        Notes
        -----
        The effect is achieved by multiplying the array's data by the same power of 2
        that would be required to bring the `abs_tolerance` to between 1 and 2.  Thus,
        all digits less significant than 1 are less significant than `abs_tolerance` —
        meaning that we can apply the standard `round` routine to set these digits to
        0.  We then divide by that same power of 2 to bring the array data back to
        nearly its original value.  By working with powers of 2, we ensure that the 0s
        at the intermediate stage are represented as 0 bits in the final result.

        For floats and array-like objects, all values must be strictly positive, or
        `inf` or `nan` will result.

        """
        if callable(abs_tolerance):
            abs_tolerance = abs_tolerance(self)
        power_of_2 = (2.0 ** np.floor(-np.log2(abs_tolerance)))
        ndarray = self.ndarray
        ndarray *= power_of_2
        np.round(ndarray, out=ndarray)
        ndarray /= power_of_2



"""Simple greedy algorithm from NINJA2 project"""

def minimal_grid(t, amp, phi, amp_tol=1e-5, phi_tol=1e-5):
    """Select a minimal grid from the input data allowing linear interpolation to given tolerances

    Storing the input data on the output grid will allow linear interpolation that
    agrees with the input data to within the given tolerances.

    This function is essentially a direct translation of the one originally written
    as part of the NINJA2 project <https://arxiv.org/abs/0709.0093>, which was
    provided as a utility named MinimizeGrid in the utilities/GridRefinement
    directory of the ninja2 svn repository.

    """
    import numpy as np

    # The objective here will be to create a vector of `bool`s, the same length as
    # t.  The truth value will correspond to whether or not that time step should
    # be kept in the final data.  We begin by assuming that the very first and last
    # steps should obviously be kept.  Then, there are two stages.  First is a
    # coarse stage, which steps through the data making intervals small enough to
    # reproduce the phi data at the interval's midpoint to within phi_tol, but no
    # smaller.  Second is the finer stage, which goes through each interval,
    # checking that every single point in the input data can be reproduced to
    # within phi_tol and amp_tol.  If that's not true, the interval is split evenly
    # into two, and the algorithm proceeds with the earlier interval.

    i0 = 0
    i1 = ((t.size-1) >> 1)  # = midpoint of the input data set
    n_points = 2
    include_indices = np.zeros(t.size(), dtype=bool)
    include_indices[0] = True
    include_indices[-1] = True

    # Returns True if the input (x,y) data can be interpolated between indices i0
    # and i1 to within a tolerance of tol at the midpoint between i0 and i1.
    def check(x, y, i0, i1, tol):
        if i1==i0+1:
            return True
        i_mid = ((i0+i1) >> 1)  # = floor(avg(i0+i1))
        if abs(y[i0] + (x[i_mid]-x[i0])*(y[i1]-y[i0])/(x[i1]-x[i0]) - y[i_mid]) > tol:
            return False  # Interval is not fine enough
        return True  # Interval is fine enough

    # Given data (t,phi), some initial index i0, and a guess for i1, this function
    # outputs the optimal index i1 so that (t,phi) can be interpolated between i0
    # and i1 to just within phi_tol of the full input data set.  Compare Numerical
    # Recipes's 'hunt' function; this is basically a hunt for that optimal index.
    def hunt(t, phi, phi_tol, i0, i1):
        inc = 1
        i1lo = i1
        i1hi = i1+1

        # Bracket the optimal i1 between i1lo and i1hi
        if check(t, phi, i0, i1lo, phi_tol):
            while check(t, phi, i0, i1hi, phi_tol):
                i1lo = i1hi;
                inc *= 2;
                i1hi += inc;
                if i1hi>int(t.size):
                    i1hi = t.size
                    break
        else:
            if i1lo<=i0+2:
                return i0+2
            i1hi = i1lo
            i1lo -= 1
            while not check(t, phi, i0, i1lo, phi_tol):
                i1hi = i1lo
                inc *= 2
                if inc > i1hi:
                    i1lo = i0+2
                    break
                else:
                    i1lo = i1hi-inc

        # Now use bisection between i1lo and i1hi
        while i1hi-i1lo != 1:
            i1m = ((i1hi+i1lo)>>1)
            if check(t, phi, i0, i1m, phi_tol):
                i1lo = i1m
            else:
                i1hi = i1m

        return i1lo

    # Coarse -- check only phi at midpoints of each interval This loop starts from
    #   the beginning of the data set, and forms the smallest interval such that
    #   the phi tolerance is achieved by linear interpolation.  Then, it moves to
    #   the end of that interval to find the next interval, etc.
    while ((i0+i1)>>1) < include_indices.size-1:
        # hunt for optimal i1
        i1 = hunt(t, phi, phi_tol, i0, i1)

        if not include_indices[i1]:
            include_indices[i1] = True
            n_points += 1
        i0 = i1
        i1 = 2*i1 - i0
        if i1<i0+2:
            i1 = i0+2

    # Fine -- check amp and phi at every point This loop goes through each of the
    #   intervals found above, and makes sure that every data point in both phi and
    #   amp can be reconstructed to within the given tolerances.  If not, it just
    #   adds the midpoint of the interval, and checks the new interval again.
    i0 = 0
    i1 = 1
    i = 1
    while i<t.size:
        if i>i1:  # This could happen below
            while i>i1:
                i1 += 1
            i0 = i
            while not include_indices[i0]:
                i0 -= 1
        while not include_indices[i1]:
            i1 += 1
        if i != i0 and i != i1:
            if (abs(1-(amp[i0]+(t[i]-t[i0])*(amp[i1]-amp[i0])/(t[i1]-t[i0]))/amp[i]) > amp_tol
                or abs(phi[i0]+(t[i]-t[i0])*(phi[i1]-phi[i0])/(t[i1]-t[i0])-phi[i]) > phi_tol):
                i1 = ((i0+i1)>>1)
                if not include_indices[i1]:
                    # if(i1==i0+1) { Caution(amp_tol, phi_tol); }
                    include_indices[i1] = True
                    n_points += 1
                continue
        if i==i1:
            i0 = i1
            i1 += 1
        i += 1

    return include_indices



"""Utilities for bitwise operations"""

import functools
import numpy as np
import numba as nb
from . import guvectorize


_uint_ftylists = [(nb.uint8[:], nb.uint8[:]), (nb.uint16[:], nb.uint16[:]), (nb.uint32[:], nb.uint32[:]), (nb.uint64[:], nb.uint64[:])]


@guvectorize(_uint_ftylists, '(n)->(n)')
def xor_forward(a_in, a_out):
    """Progressively XOR data along an axis

    This is primarily a helper function for `xor`.  See that function's docstring
    for details.

    """
    a_out[0] = a_in[0]
    for i in range(a_in.shape[0] - 1, 0, -1):
        a_out[i] = np.bitwise_xor(a_in[i - 1], a_in[i])


@guvectorize(_uint_ftylists, '(n)->(n)')
def xor_reverse(a_in, a_out):
    """Progressively reverse XOR data along an axis

    This is primarily a helper function for `xor`.  See that function's docstring
    for details.

    """
    a_out[0] = a_in[0]
    for i in range(1, a_in.shape[0]):
        a_out[i] = np.bitwise_xor(a_out[i - 1], a_in[i])


del _uint_ftylists


def xor(x, reverse=False, preserve_dtype=False, axis=-1, **kwargs):
    """Progressively XOR data along an axis

    This function steps through an array, starting with the second element, and
    evaluates bitwise XOR on that element and the preceding one.  This is a useful
    step in compressing reasonably continuous data.

    Parameters
    ----------
    x : array_like
        After conversion to an array, this must have a dtype that can be converted
        to unsigned integers with 1, 2, 4, or 8 bytes.
    reverse : bool, optional
        If True, this reverses the transformation of progressively XOR-ing the
        data.  In particular, `xor(xor(x), reverse=True)` will be bitwise-identical
        to `x`.
    preserve_dtype : bool, optional
        If True, the returned array will have the same dtype as the input array.
        Note that this could result in invalid data.  For example, if the input
        data has some `float` dtype, it is possible for bitwise XOR to result in
        NaNs.  The bits are usually retained correctly nonetheless, but some
        operations may produce errors or incorrect results when NaNs are present.
        If this parameter is False, the returned array will always have unsigned
        integer type, of a bit width equal to the input array.
    axis : int, optional
        This is the axis along which the operation will be performed.  You probably
        want this to correspond to the most quickly varying dimension of the data —
        the time axis of a timeseries, for example.

    Returns
    -------
    u : ndarray
        Depending on `preserve_dtype`, this will either be an unsigned integer
        array (the default) or will have the same dtype as the input data.

    """
    x = np.asarray(x)
    itemsize = x.itemsize
    if itemsize not in [1, 2, 4, 8]:
        raise ValueError(f"Input array's byte size must be one of {{1, 2, 4, 8}}, not {itemsize}")
    dtype = np.dtype(f"u{itemsize}")
    u = x.view(dtype)
    kwargs["axis"] = axis
    if "out" in kwargs:
        kwargs["out"] = np.asarray(kwargs["out"]).view(dtype)
    if reverse:
        u_xor = xor_reverse(u, **kwargs)
    else:
        u_xor = xor_forward(u, **kwargs)
    if preserve_dtype:
        return u_xor.view(x.dtype)
    else:
        return u_xor


@functools.lru_cache()
def multishuffle(shuffle_widths, forward=True):
    """Construct functions to "multi-shuffle" data

    This is a generalization of the `shuffle` algorithm found in HDF5, or the
    `bitshuffle` algorithm — and in some sense interpolates between them.  See
    below for more explanation.

    Parameters
    ----------
    shuffle_widths : array_like
        This must be an iterable of integers, representing the number of bits in
        each piece of each number that is shuffled, starting from the highest
        significance, and proceeding to the lowest.  The sum of these numbers must
        be the total bit width of the numbers that will be given as input — which
        must currently be 8, 16, 32, or 64.  There is no restriction on the
        individual widths, but note that if they do not fit evenly into 8-bit
        bytes, the result is unlikely to compress well.
    forward : bool, optional
        If True (the default), the returned function will shuffle data; if False,
        the returned function will reverse this process — unshuffle.

    Returns
    -------
    shuffle_func : function
        This function takes just one parameter — the array to be shuffled — and
        returns the shuffled array.  Note that the input array *must* be flat (have
        just one dimension), and will be viewed as an array of unsigned integers of
        the input bit width.  This can affect the shape of the array and order of
        elements.  You should ensure that this process will result in an array of
        numbers in the order that you want.  For example, if you have a 2-d array
        of floats `a` that are more continuous along the first dimension, you might
        pass `np.ravel(a.view(np.uint64), 'F')`, where F represents Fortran order,
        which varies more quickly in the first dimension.  Also note that this
        function is JIT-compiled by numba.

    Notes
    -----
    The standard "shuffle" algorithm (as found in HDF5, for example) takes an array
    of numbers and shuffles their bytes so that all bytes of a given significance
    are stored together — the first byte of all the numbers are stored
    contiguously, then the second byte of all the numbers, and so on.  The
    motivation for this is that — with reasonably smooth data — bytes in the same
    byte position in sequential numbers are usually more related to each other than
    they are to other bytes within the same number, which means that shuffling
    results in better compression of the data.

    There is no reason that shuffling can only work byte-by-byte, however.  There
    is also a "bitshuffle" algorithm, which works in the same way, but collecting
    bits rather than bytes.  More generally, we could vary the number of bits
    stored together as we move along the numbers.  For example, we could store the
    first 8 bits of each number, followed by the next 4 bits of each number, etc.
    This is the "multi-shuffle" algorithm.

    With certain types of data, this can reduce the compressed data size
    significantly.  For example, with float data for which successive values have
    been XOR-ed, the sign bit will very rarely change, the next 11 bits
    (representing the exponent) and a few of the following bits (representing the
    highest-significance digits) will typically be highly correlated, while as we
    move to lower significance there will be less correlation.  Thus, we might
    shuffle the first 8 bits together, followed by the next 8, then the next 4, the
    next 4, the next 2, and so on — decreasing the shuffle width as we go.  The
    `shuffle_widths` input might look like [8, 8, 4, 4, 2, 2, 1, 1, 1, 1, ...].

    There are also some cases where we see correlation *increasing* again at low
    significance.  For example, if a number results from cancellation — the
    subtraction of two numbers much larger than their difference — then its
    lower-significance bits will be 0.  If we then multiply that by some integer
    (e.g., for normalization), there may be some very correlated but nonzero
    pattern.  In either case, compression might improve if the values at the end of
    our shuffle_widths list increase.

    """
    import numpy as np
    from . import jit

    bit_width = np.sum(shuffle_widths, dtype=np.int64)  # uint64 casts to float under floor division...
    if bit_width not in [8, 16, 32, 64]:
        raise ValueError(f"Total bit width must be one of [8, 16, 32, 64], not {bit_width}")
    dtype = np.dtype(f"u{bit_width//8}")
    bit_width = dtype.type(bit_width)
    reversed_shuffle_widths = np.array(list(reversed(shuffle_widths)), dtype=dtype)
    one = dtype.type(1)

    if forward:

        def shuffle(a):
            a = a.view(dtype)
            if a.ndim != 1:
                raise ValueError(
                    "\nThis function only accepts flat arrays.  Make sure you flatten "
                    "(using ravel, reshape, or flatten)\n in a way that keeps your data"
                    "contiguous in the order you want."
                )
            b = np.zeros_like(a)
            b_array_bit = np.uint64(0)
            for i, shuffle_width in enumerate(reversed_shuffle_widths):
                mask_shift = np.sum(reversed_shuffle_widths[:i])
                mask = dtype.type(2 ** shuffle_width - 1)
                pieces_per_element = bit_width // shuffle_width
                for a_array_index in range(a.size):
                    b_array_index = b_array_bit // bit_width
                    b_element_bit = dtype.type(b_array_bit % bit_width)
                    masked = (a[a_array_index] >> mask_shift) & mask
                    b[b_array_index] += masked << b_element_bit
                    if b_element_bit + shuffle_width > bit_width:
                        b[b_array_index + one] += masked >> (bit_width - b_element_bit)
                    b_array_bit += shuffle_width
            return b

        return jit(shuffle)

    else:
        # This function is almost the same as above, except for:
        # 1) swap a <-> b in input and output
        # 2) reverse the effect of the line in which b was set from a
        def unshuffle(b):
            b = b.view(dtype)
            if b.ndim != 1:
                raise ValueError(
                    "\nThis function only accepts flat arrays.  Make sure you flatten "
                    "(using ravel, reshape, or flatten)\n in a way that keeps your data"
                    "contiguous in the order you want."
                )
            a = np.zeros_like(b)
            b_array_bit = dtype.type(0)
            for i, shuffle_width in enumerate(reversed_shuffle_widths):
                mask_shift = np.sum(reversed_shuffle_widths[:i])
                mask = dtype.type(2 ** shuffle_width - 1)
                pieces_per_element = bit_width // shuffle_width
                for a_array_index in range(a.size):
                    b_array_index = b_array_bit // bit_width
                    b_element_bit = b_array_bit % bit_width
                    masked = (b[b_array_index] >> b_element_bit) & mask
                    a[a_array_index] += masked << mask_shift
                    if b_element_bit + shuffle_width > bit_width:
                        a[a_array_index] += (
                            (b[b_array_index + one] << (bit_width - b_element_bit)) & mask
                        ) << mask_shift
                    b_array_bit += shuffle_width
            return a

        return jit(unshuffle)



"""Compute contributions to memory in asymptotic waveforms

This code is based on the paper "Adding Gravitational Memory to Waveform
Catalogs using BMS Balance Laws" by Mitman et al.  The main result of that
paper is encapsulated in the `add_memory` function.  Basic usage looks like
this:

```python
h = sxs.load("SXS:BBH:0123/Lev/rhOverM", extrapolation_order=3)
h_with_memory = sxs.waveforms.memory.add_memory(h, integration_start_time=1000.0)
```

"""

import numpy as np
import spherical
from . import WaveformModes


class _ModesTimeSeries(WaveformModes):

    @property
    def s(self):
        return self.spin_weight

    @property
    def u(self):
        return self.time

    from spherical.modes.algebra import (
        conjugate, bar, _real_func, real, _imag_func, imag, norm,
        add, subtract, multiply, divide
    )

    conj = conjugate

    from spherical.modes.utilities import (
        truncate_ell, _check_broadcasting
    )

    from spherical.modes.ufuncs import __array_ufunc__


def MTS(*args, **kwargs):
    kwargs.setdefault('multiplication_truncator', max)
    return _ModesTimeSeries(*args, **kwargs)


def 𝔇(h_mts):
    """Differential operator 𝔇 acting on spin-weight s=0 function

    As defined in Eq. (7b) of 'Adding Gravitational Memory to Waveform Catalogs
    using BMS Balance Laws'

    """
    h = h_mts.copy()
    s = h.ndarray

    for ell in range(h.ell_min, h.ell_max + 1):
        if ell < 2:
            𝔇_value = 0
        else:
            𝔇_value = 0.125 * (ell + 2) * (ell + 1) * (ell) * (ell - 1)
        s[..., h.index(ell, -ell) : h.index(ell, ell) + 1] *= 𝔇_value

    return h


def 𝔇inverse(h_mts):
    """Inverse of differential operator 𝔇 acting on spin-weight s=0 function

    As defined in Eq. (7b) of 'Adding Gravitational Memory to Waveform Catalogs
    using BMS Balance Laws'

    """
    h = h_mts.copy()
    s = h.ndarray

    for ell in range(h.ell_min, h.ell_max + 1):
        if ell < 2:
            𝔇inverse_value = 0
        else:
            𝔇inverse_value = 1.0 / (0.125 * (ell + 2) * (ell + 1) * (ell) * (ell - 1))
        s[..., h.index(ell, -ell) : h.index(ell, ell) + 1] *= 𝔇inverse_value

    return h


def 𝔇inverseLaplacianinverse(h_mts):
    """Inverse of differential operator D² 𝔇 acting on spin-weight s=0 function

    As defined in Eqs. (7) of 'Adding Gravitational Memory to Waveform Catalogs
    using BMS Balance Laws'

    """
    h = h_mts.copy()
    s = h.ndarray

    for ell in range(h.ell_min, h.ell_max + 1):
        if ell < 2:
            𝔇inverse_value = 0
        else:
            𝔇inverse_value = -1.0 / (0.125 * (ell + 2) * (ell + 1)**2 * (ell)**2 * (ell - 1))
        s[..., h.index(ell, -ell) : h.index(ell, ell) + 1] *= 𝔇inverse_value

    return h


def mass_aspect(Psi2, h):
    h = MTS(h)
    Psi2 = MTS(Psi2)
    return - (Psi2 + 0.25 * h.dot * h.bar).re


def J_m(h, Psi2):
    """Bondi mass aspect contribution to electric part of strain

    Calculated according to Eq. (17a) of 'Adding Gravitational Memory to Waveform
    Catalogs using BMS Balance Laws'

    Parameters
    ----------
    h : WaveformModes
        WaveformModes object corresponding to the strain
    Psi2 : WaveformModes
        WaveformModes object corresponding to Psi2

    Returns
    -------
    J_m : WaveformModes
        Bondi mass aspect contribution to the strain

    """
    h = MTS(h)
    Psi2 = MTS(Psi2)
    m = mass_aspect(Psi2, h)
    J_m = 0.5 * 𝔇inverse(m).ethbar.ethbar

    return J_m


def J_E(h, integration_start_time=None):
    """Energy flux contribution to electric part of strain

    Calculated according to Eq. (17b) of 'Adding Gravitational Memory to Waveform
    Catalogs using BMS Balance Laws'

    Parameters
    ----------
    h : WaveformModes
        WaveformModes object corresponding to the strain
    integration_start_time : float, optional
        The time at which the energy flux integral should begin.  The default is
        `h.t[0]`.

    Returns
    -------
    J_E : WaveformModes
        Energy flux contribution to the strain

    """

    hdot = MTS(h).dot

    J_ℰ = 0.5 * 𝔇inverse(0.25 * (hdot * hdot.bar).int).ethbar.ethbar

    if integration_start_time is not None:
        integration_start_time_index = np.argmin(abs(h.t - integration_start_time))
        J_ℰ -= J_ℰ[integration_start_time_index, :]

    return J_ℰ


def J_Nhat(h, Psi2):
    """Angular momentum aspect contribution to magnetic part of strain

    Calculated according to Eq. (17c) of 'Adding Gravitational Memory to Waveform
    Catalogs using BMS Balance Laws'

    Parameters
    ----------
    h : WaveformModes
        WaveformModes object corresponding to the strain
    Psi2 : WaveformModes
        WaveformModes object corresponding to Psi2

    Returns
    -------
    J_Nhat : WaveformModes
        Bondi angular momentum aspect contribution to the strain

    """

    h = MTS(h)
    Psi2 = MTS(Psi2)

    # # Note that the contributions from the last two terms drop out as soon as we
    # # take the imaginary part below.
    # m = mass_aspect(Psi2, h)
    # N̂ = 2 * Psi1 - 0.25 * (h.bar * h.eth) - h.t * m.eth - 0.125 * (h * h.bar).eth

    # Ψ̇₁ = - ½ (ðΨ₂ - ½ h̄ ðḣ)
    # N̂̇ = 2 Ψ̇₁ - ¼ h̄ ðḣ - ¼ h̄̇ ðh
    #   = -ðΨ₂ + ½ h̄ ðḣ - ¼ h̄ ðḣ - ¼ h̄̇ ðh
    #   = -ðΨ₂ + ¼ h̄ ðḣ - ¼ h̄̇ ðh

    N̂dot = -Psi2.eth + 0.25 * (h.bar * h.dot.eth - h.bar.dot * h.eth)
    J_N̂ = 0.5j * 𝔇inverseLaplacianinverse(N̂dot.ethbar.im).ethbar.ethbar

    return J_N̂


def J_J(h):
    """Angular momentum flux contribution to magnetic part of strain

    Calculated according to Eq. (17d) of 'Adding Gravitational Memory to Waveform
    Catalogs using BMS Balance Laws'

    Parameters
    ----------
    h : WaveformModes
        WaveformModes object corresponding to the strain

    Returns
    -------
    J_J : WaveformModes
        Angular momentum flux contribution to the strain

    """

    h = MTS(h)
    hdot = h.dot
    J_𝒥 = 0.5j * 𝔇inverseLaplacianinverse(
        0.125 * (3 * h * hdot.bar.ethbar - 3 * hdot * h.bar.ethbar + hdot.bar * h.ethbar - h.bar * hdot.ethbar).eth.im
    ).ethbar.ethbar

    return J_𝒥


def add_memory(h, integration_start_time=None):
    """Add electric component of null memory to strain

    This adds the contribution from the energy flux to the strain.

    Parameters
    ----------
    h : WaveformModes
        WaveformModes object corresponding to the strain
    integration_start_time : float, optional
        Time at which the energy flux integral should begin.  The default is
        `h.t[0]`.

    Returns
    -------
    h_with_memory : WaveformModes
        WaveformModes object corresponding to the strain with electric memory

    """
    h_with_memory = WaveformModes(MTS(h) + J_E(h, integration_start_time=integration_start_time))
    h_with_memory.register_modification(add_memory, integration_start_time=integration_start_time)
    return h_with_memory



"""Collection of functions to reduce data size

This submodule provides several related, though not quite directly compatible,
routines to reduce the number of points in a data set while retaining a given
level of accuracy:

* The original NINJA2 routine in `linear_bisection`
* A routine emulating `romspline`'s in `greedy_spline`
* A new routine in `peak_greed`

The original NINJA2 routine was limited by the fact that it had to be written
as a single easy-to-understand and easy-to-compile C file.  As a result, it
used linear interpolation, rather than the more reasonable cubic spline.  It is
included here for historical reasons, but probably should not generally be
used.

The `romspline` version is very slow, because the algorithm essentially scales
as N**2, where N is the number of points in the final output.  This is because
one point is added per iteration, each iteration requires constructing a new
spline for all the current points, and the splines dominate the timing.

The `peak_greed` routine gets around this problem by selecting multiple new
points on each iteration, guided by the peaks of the error quantity.  The
result has nearly the same number of points, but the timing scales with N,
typically providing results dozens to hundreds of times more quickly.

"""

from .suppression import suppressor, suppress



known_creators = {
    'Michael Boyle': {
        'name': 'Boyle, Michael',
        'affiliation': 'Cornell University',
        'orcid': '0000-0002-5075-5116',
    },
    'Mike Boyle': {
        'name': 'Boyle, Michael',
        'affiliation': 'Cornell University',
        'orcid': '0000-0002-5075-5116',
    },
    'Luisa Buchman': {
        'name': 'Buchman, Luisa',
        'affiliation': 'Caltech',
    },
    'Matthew Duez': {
        'name': 'Duez, Matthew',
        'affiliation': 'Washington State',
    },
    'Francois Foucart': {
        'name': 'Foucart, Francois',
        'affiliation': 'University of New Hampshire',
    },
    'Michael Grudich': {
        'name': 'Grudich, Michael',
        'affiliation': 'Caltech',
    },
    'Dan Hemberger': {
        'name': 'Hemberger, Dan',
    },
    'Larry Kidder': {
        'name': 'Kidder, Larry',
        'affiliation': 'Cornell University',
    },
    'Geoffrey Lovelace': {
        'name': 'Lovelace, Geoffrey',
        'affiliation': 'Cal State Fullerton',
        'orcid': '0000-0002-7084-1070',
    },
    'Ilana MacDonald': {
        'name': 'MacDonald, Ilana',
    },
    'Abdul Mroue': {
        'name': 'Mroue, Abdul',
    },
    'Harald Pfeiffer': {
        'name': 'Pfeiffer, Harald',
        'affiliation': 'AEI Potsdam',
        'orcid': '0000-0001-9288-519X',
    },
    'Mark Scheel': {
        'name': 'Scheel, Mark',
        'affiliation': 'Caltech',
    },
    'Bela Szilagyi': {
        'name': 'Szilagyi, Bela',
    },
    'Nicholas Taylor': {
        'name': 'Taylor, Nicholas',
    },
    'Anil Zenginoglu': {
        'name': 'Zenginoglu, Anil',
    }
}

creators_emails = {
    'mgiesler@tapir.caltech.edu': {'name': 'Matt Giesler'},
    # 'glovelace@fullerton.edu': { 'name': 'Geoffrey Lovelace' },  # in the default list, so don't worry about it
    'kevinbarkett@yahoo.com': {'name': 'Kevin Barkett'},
    'jonathan.blackman.0@gmail.com': {'name': 'Jonathan Blackman'},
    'ian.hinder@aei.mpg.de': {'name': 'Ian Hinder'},
    # 'pfeiffer@cita.utoronto.ca': { 'name': 'Harald Pfeiffer' },  # in the default list, so don't worry about it
    'tonyc@cita.utoronto.ca': {'name': 'Tony Chu'},
    'kbarkett@caltech.edu': {'name': 'Kevin Barkett'},
    'serguei.ossokine@aei.mpg.de': {'name': 'Serguei Ossokine'},
    'ossokine@aei.mpg.de': {'name': 'Serguei Ossokine'},
}

default_creators = [
    known_creators[person]
    for person in [
        'Larry Kidder', 'Harald Pfeiffer', 'Mark Scheel',
        'Michael Boyle', 'Dan Hemberger', 'Geoffrey Lovelace', 'Bela Szilagyi'
    ]
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sxs',
 'sxs.catalog',
 'sxs.horizons',
 'sxs.metadata',
 'sxs.utilities',
 'sxs.utilities.decimation',
 'sxs.utilities.lvcnr',
 'sxs.utilities.references',
 'sxs.waveforms',
 'sxs.zenodo',
 'sxs.zenodo.api']

package_data = \
{'': ['*']}

install_requires = \
['feedparser>=6.0.1,<7.0.0',
 'h5py>=3,<4',
 'inflection>=0.5.1,<0.6.0',
 'numpy>=1.20,<2.0',
 'pandas>=1.1.2,<2.0.0',
 'pylatexenc>=2.7,<3.0',
 'pytest-forked>=1.3.0,<2.0.0',
 'pytz>=2020.1,<2021.0',
 'quaternionic>=1.0,<2.0',
 'requests>=2.24.0,<3.0.0',
 'scipy>=1.0,<2.0',
 'spherical>=1.0,<2.0',
 'tqdm>=4.48.2,<4.61.2',
 'urllib3>=1.25.10,<2.0.0']

extras_require = \
{':implementation_name == "cpython"': ['numba>=0.55'],
 'ecosystem': ['ipywidgets>=7.5.1,<8.0.0',
               'ipykernel>=5.3.4,<6.0.0',
               'jupyter_contrib_nbextensions>=0.5.1,<0.6.0',
               'jupyterlab>=2.2.8,<3.0.0',
               'line_profiler>=3.0.2,<4.0.0',
               'memory_profiler>=0.57.0,<0.58.0',
               'matplotlib>=2.1.1',
               'sympy>=1.6.2,<2.0.0',
               'corner>=2.1.0,<3.0.0',
               'qgrid>=1.3.1,<2.0.0',
               'rise>=5.6.1,<6.0.0',
               'quaternion>=0.3.1,<0.4.0'],
 'ecosystem:sys_platform != "win32"': ['scri>=2020.8.18,<2021.0.0'],
 'mkapi:python_version >= "3.7"': ['mkapi==1.0.13'],
 'mkdocs:implementation_name == "cpython"': ['mkdocs>=1.1.2'],
 'pymdown-extensions:implementation_name == "cpython"': ['pymdown-extensions>=8,<9']}

setup_kwargs = {
    'name': 'sxs',
    'version': '2022.0.0',
    'description': 'Interface to data produced by the Simulating eXtreme Spacetimes collaboration',
    'long_description': '[![Test Status](https://github.com/sxs-collaboration/sxs/workflows/tests/badge.svg)](https://github.com/sxs-collaboration/sxs/actions)\n[![Documentation Status](https://readthedocs.org/projects/sxs/badge/?version=main)](https://sxs.readthedocs.io/en/main/?badge=main)\n[![PyPI Version](https://img.shields.io/pypi/v/sxs?color=)](https://pypi.org/project/sxs/)\n[![Conda Version](https://img.shields.io/conda/vn/conda-forge/sxs.svg?color=)](https://anaconda.org/conda-forge/sxs)\n[![MIT License](https://img.shields.io/badge/license-MIT-blue.svg)](https://github.com/sxs-collaboration/sxs/blob/main/LICENSE)\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/moble/sxs_notebooks/master)\n\n\n# Simulating eXtreme Spacetimes python package\n\nThe `sxs` python package provides a high-level interface for using data\nproduced by the SXS collaboration.  In particular, the function `sxs.load` can\nautomatically find, download, and load data, returning objects that provide\ncommon interfaces to the various types of data, without forcing the user to\nworry about details like data formats or where to find the data.  It can also\nautomatically select the newest or highest-resolution dataset for a given\nsimulation, or return a range of versions or resolutions.  Currently, the\nhigh-level objects encapsulate\n\n  * Catalog — a listing of all data produced by the SXS collaboration\n  * Metadata — data describing the simulation parameters\n  * Horizons — time-series data describing the apparent horizons\n  * Waveforms — time-series data describing the extrapolated gravitational-wave\n    modes\n\n\n## Installation\n\nBecause this package is pure python code, installation is very simple.  In\nparticular, with a reasonably modern installation, you can just run a command\nlike\n\n```bash\nconda install -c conda-forge sxs\n```\n\nor\n\n```bash\npython -m pip install sxs\n```\n\nHere, `conda` requires the [conda](https://docs.anaconda.com/anaconda/install/)\ninstallation of python, which is the most recommended approach for scientific\npython; the second command assumes that you have an appropriate python\nenvironment set up in some other way.  Either of these commands will download\nand install the `sxs` package and its most vital requirements.\n\nIf you want to install all the goodies that enable things like jupyter\nnotebooks with plots and interactive tables, you could run\n\n```bash\nconda install -c conda-forge sxs-ecosystem\n```\n\nor\n\n```bash\npython -m pip install sxs[ecosystem]\n```\n\nYou will probably also want to set some sensible defaults to automatically\ndownload and cache data:\n\n```bash\npython -c "import sxs; sxs.write_config(download=True, cache=True)"\n```\n\nThis will write a configuration file in the directory returned by\n`sxs.sxs_directory("config")`, and downloaded data will be cached in the\ndirectory returned by `sxs.sxs_directory("cache")`.  See [that function\'s\ndocumentation](api/sxs.utilities.sxs_directories/#sxsutilitiessxs_directoriessxs_directory)\nfor details.\n\n\n## Usage\n\nAn extensive demonstration of this package\'s capabilities is available\n[here](https://mybinder.org/v2/gh/moble/sxs_notebooks/master), in the form of\ninteractive jupyter notebooks that are actually running this code and some\npre-downloaded data.  The following is just a very brief overview of the `sxs`\npackage\'s main components.\n\nThere are four important objects to understand in this package:\n\n```python\nimport sxs\n\ncatalog = sxs.load("catalog")\nmetadata = sxs.load("SXS:BBH:0123/Lev/metadata.json")\nhorizons = sxs.load("SXS:BBH:0123/Lev/Horizons.h5")\nwaveform = sxs.load("SXS:BBH:0123/Lev/rhOverM", extrapolation_order=2)\n```\n\n[The `catalog` object](api/sxs.catalog.catalog/#sxs.catalog.catalog.Catalog)\ncontains information about every simulation in the catalog, including all\navailable data files, and information about how to get them.  You probably\ndon\'t need to actually know about details like where to get the data, but\n`catalog` can help you find the simulations you care about.  Most importantly,\n`catalog.simulations` is a `dict` object, where the keys are names of\nsimulations (like "SXS:BBH:0123") and the values are the same types as [the\n`metadata` object](api/sxs.metadata.metadata/#sxs.metadata.metadata.Metadata),\nwhich contains metadata about that simulation — things like mass ratio, spins,\netc.  This `metadata` reflects the actual output of the simulations, which\nleads to some inconsistencies in their formats.  A more consistent interface\n(though it is biased toward returning NaNs where a human might glean more\ninformation) is provided by `catalog.table`, which returns a\n[`pandas`](https://pandas.pydata.org/docs/) `DataFrame` with specific data\ntypes for each column.\n\nThe actual data itself is primarily contained in the next two objects.  [The\n`horizons` object](api/sxs.horizons/#sxs.horizons.Horizons) has three\nattributes — `horizons.A`, `horizons.B`, and `horizons.C` — typically\nrepresenting the original two horizons of the black-hole binary and the common\nhorizon that forms at merger.  In matter simulations, one or more of these may\nbe `None`.  Otherwise, each of these three is a\n[`HorizonQuantities`](api/sxs.horizons/#sxs.horizons.HorizonQuantities) object,\ncontaining several timeseries relating to mass, spin, and position.\n\nFinally, the\n[`waveform`](api/sxs.waveforms.waveform_modes/#sxs.waveforms.waveform_modes.WaveformModes)\nencapsulates the modes of the waveform and the corresponding time information,\nalong with relevant metadata like data type, spin weight, etc., and useful\nfeatures like numpy-array-style slicing.\n',
    'author': 'Michael Boyle',
    'author_email': 'michael.oliver.boyle@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sxs-collaboration/sxs',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
